#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
import traceback
from collections import defaultdict
from functools import cache
from typing import cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def MeshChangeElemType(mesh: meshio.Mesh) -> meshio.Mesh:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common_progress import ProgressBar
    from pyhope.mesh.mesh_vars import ELEMTYPE, nGeo
    # ------------------------------------------------------

    # Split hexahedral elements if requested
    nZones    = mesh_vars.nZones
    elemTypes = mesh_vars.elemTypes
    elemNames = ['' for _ in range(nZones)]  # noqa: E271

    # No element types given
    if len(elemTypes) == 0:
        return mesh

    # Fully hexahedral mesh
    if all(elemType % 100 == 8 for elemType in elemTypes):
        return mesh

    # Simplex elements requested
    if any(elemType % 100 != 8 for elemType in elemTypes):
        if mesh_vars.nGeo > 4:
            hopout.warning('Non-hexahedral elements are not supported for nGeo > 4, exiting...')
            sys.exit(1)

    hopout.info('Converting hexahedral elements to simplex elements')

    # Instantiate ELEMTYPE
    elemTypeInam = ELEMTYPE().inam

    for i in range(nZones):
        if nGeo == 1:
            elemNames[i] = elemTypeInam[elemTypes[i]][0]
        else:
            # check whether user entered correct high-order element type
            if elemTypes[i] < 200:
                # Adapt to high-order element type
                elemTypes[i] += 100

            # Get the element name and skip the entries for incomplete 2nd order elements
            try:
                if elemTypes[i] % 100 == 5:    # pyramids (skip 1)
                    elemNames[i] = elemTypeInam[elemTypes[i]][1]
                elif elemTypes[i] % 100 == 6:  # prisms (skip 1)
                    elemNames[i] = elemTypeInam[elemTypes[i]][nGeo-1]
                elif elemTypes[i] % 100 == 8:  # hexahedra (skip 2)
                    elemNames[i] = elemTypeInam[elemTypes[i]][nGeo]
                else:                          # tetrahedra
                    elemNames[i] = elemTypeInam[elemTypes[i]][nGeo-2]
            except IndexError:
                hopout.warning('Element type {} not supported for nGeo = {}, exiting...'.format(elemTypes[i], nGeo))
                sys.exit(1)

    # Copy original points
    pointl    = cast(list, mesh.points.tolist())
    elems_old = mesh.cells.copy()
    cell_sets = getattr(mesh, 'cell_sets', {})

    # Get base key to distinguish between linear and high-order elements
    ho_key = 100 if nGeo == 1 else 200

    # Set up the element splitting function
    elemSplitter = {ho_key + 4: (split_hex_to_tets , tetra_faces),
                    ho_key + 5: (split_hex_to_pyram, pyram_faces),
                    ho_key + 6: (split_hex_to_prism, prism_faces),
                    # Keep hexahedral elements as they are
                    ho_key + 8: (split_hex_to_hex  , hex_faces  )}

    faceMaper = { ho_key + 4: lambda x: 0,
                  ho_key + 5: lambda x: 0 if x == 0 else 1,
                  ho_key + 6: lambda x: 0 if x == 0 else 1,
                  # Keep hexahedral elements as they are
                  ho_key + 8: lambda x: 1}
    nFace   = (nGeo+1)*(nGeo+2)/2

    # Convert the (quad) boundary cell set into a dictionary
    csets_old = {}

    for cname, cblock in cell_sets.items():
        # Each set_blocks is a list of arrays, one entry per cell block
        for blockID, block in enumerate(cblock):
            if elems_old[blockID].type[:4] != 'quad':
                continue

            # Ignore the volume zones
            if block is None:
                continue

            # Sort them as a set for membership checks
            for face in block:
                nodes = mesh.cells_dict[elems_old[blockID].type][face]
                csets_old.setdefault(frozenset(nodes), []).append(cname)

    nPoints  = len(pointl)
    nFaces   = np.zeros(2, dtype=int)
    match nGeo:
        case 1:
            faceType = ['triangle'  , 'quad'  ]
            faceNum  = [          3 ,       4 ]
        case 2:
            faceType = ['triangle6' , 'quad9' ]
            faceNum  = [          6 ,       9 ]
        # case 3:
        #     faceType = ['triangle10', 'quad16']
        #     faceNum  = [         10 ,      16 ]
        case 4:
            faceType = ['triangle15', 'quad25']
            faceNum  = [         15 ,      25 ]
        case _:
            hopout.warning('nGeo = {} not supported for element splitting'.format(nGeo))
            sys.exit(1)

    # Prepare new cell blocks and new cell_sets
    elems_lst = {ftype: [] for ftype in faceType}
    csets_lst = {}

    # Create the element sets
    meshcells = tuple((k, v) for k, v in mesh.cell_sets_dict.items() if any(key.startswith('hexahedron') for key in v.keys()))

    # If meshcells is empty, we fake it assign it to Zone1
    if len(meshcells) == 0:
        meshcells = tuple(('Zone1', np.array([i for i in range(len(v))])) for k, v in mesh.cells_dict.items()
                                                                                   if k.startswith('hexahedron'))

    nTotalElems = sum(cdata.shape[0] for _, zdata in meshcells for _, cdata in cast(dict, zdata).items())
    bar = ProgressBar(value=nTotalElems, title='│             Processing Elements', length=33, threshold=1000)

    # Build an inverted index to map each node to all face keys (from csets_old) that contain it
    nodeToFace = defaultdict(set)
    for subFace in csets_old:
        for node in subFace:
            nodeToFace[node].add(subFace)

    for iElem, meshcell in enumerate(meshcells):
        _    , mdict = meshcell
        mtype, mcell = list(cast(dict, mdict).keys())[0], list(cast(dict, mdict).values())[0]

        elemType     = elemTypes[iElem]
        elemName     = elemNames[iElem]

        split, faces = elemSplitter.get(elemType, (None, None))
        faceMap      = faceMaper.get(elemType, None)
        # Sanity check
        if faceMap is None:
            sys.exit(1)

        cdata = mesh.get_cells_type(mtype)[mcell]

        if split is None or faces is None:
            hopout.warning('Element type {} not supported for splitting'.format(elemTypes[iElem]))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)

        elemSplit = split(nGeo)

        # Hex block: Iterate over each element
        for elem in cdata:
            # Pyramids need a center node
            if elemType % 100 == 5:
                # Find the element orientation
                # > The first 8 points (indices) in elem are the CGNS-ordered vertices
                vertices = np.array([pointl[i] for i in elem[:8]])

                # v[0] is origin, v[1] is local x-direction, v[2] is local y-direction, v[3] is local z-direction
                # > This only works for trapezoidal elements
                v        = [vertices[0], vertices[1]-vertices[0], vertices[3]-vertices[0], vertices[4]-vertices[0]]

                # Compute the element center
                center   = np.mean(np.array([pointl[i] for i in elem]), axis=0)

                match nGeo:
                    case 1:
                        # Append the new point to the point list
                        pointl.append(center.tolist())

                        # Overwrite the element with the new indices
                        elem     = np.array(list(elem) + [nPoints])
                        nPoints += 1
                    case 2:
                        # Generate the grid of new points
                        signarr  = [-0.5, 0.5]
                        # Create a 3D grid of sign factors

                        grid = np.array(np.meshgrid(signarr, signarr, signarr)).T.reshape(-1, 3)
                        # Combine the grid with the directional vectors v[1], v[2], v[3]
                        edge_array = center + 0.5 * (grid[:, 0][:, None] * v[1] +
                                                     grid[:, 1][:, None] * v[2] +
                                                     grid[:, 2][:, None] * v[3])
                        # Convert all points to list in one go
                        edges   = edge_array.tolist()
                        # Append the new points to the point list
                        pointl.extend(edges)

                        # Overwrite the element with the new indices
                        elem     = np.array(list(elem) + list(range(nPoints, nPoints+edge_array.shape[0])))
                        nPoints += edge_array.shape[0]
                    case 4:
                        # Generate the grid of new points
                        signarr = [-3./4., -1./4., 1./4., 3./4.]
                        # Create a 3D grid of sign factors
                        grid = np.array(np.meshgrid(signarr, signarr, signarr)).T.reshape(-1, 3)
                        # Combine the grid with the directional vectors v[1], v[2], v[3]
                        edge_array = center + 0.5 * (grid[:, 0][:, None] * v[1] +
                                                     grid[:, 1][:, None] * v[2] +
                                                     grid[:, 2][:, None] * v[3])
                        # Convert all points to list in one go
                        edges   = edge_array.tolist()
                        # Append the new points to the point list
                        pointl.extend(edges)

                        # Overwrite the element with the new indices
                        elem     = np.array(list(elem) + list(range(nPoints, nPoints+edge_array.shape[0])))
                        nPoints += edge_array.shape[0]

            # Split each element into sub-elements
            subElems = elem[elemSplit]

            for subElem in subElems:
                subFaces = tuple(np.array(subElem)[face] for face in faces(nGeo))

                for subFace in subFaces:
                    faceVal = faceMap(0) if len(subFace) == nFace else faceMap(1)
                    faceSet = frozenset(subFace)

                    # Get candidate cset keys using the nodes in the face
                    candidate_sets = [nodeToFace[node] for node in faceSet if node in nodeToFace]
                    if not candidate_sets:
                        continue

                    common_candidates = set.intersection(*candidate_sets)
                    for candidate in common_candidates:
                        # Check if the subFace is indeed a subset of the candidate from csets_old
                        if faceSet.issubset(candidate):
                            # Use the associated boundary name
                            # (Assuming all boundary names are stored in a list for this candidate. Adjust if needed.)
                            names = csets_old[candidate]
                            # Update csets_lst for each name in the list.
                            for name in names:
                                csets_lst.setdefault(name, [[], []])
                                csets_lst[name][faceVal].append(nFaces[faceVal])

                    elems_lst[faceType[faceVal]].append(np.array(subFace, dtype=int))
                    nFaces[faceVal] += 1

            elems_lst.setdefault(elemName, []).extend(subElems)

            # Update the progress bar
            bar.step()

    # Close the progress bar
    bar.close()

    # Convert lists to NumPy arrays for elems_new and csets_new
    elems_new = {}
    csets_new = {}

    for key in elems_lst:
        if   isinstance(elems_lst[key], list) and     elems_lst[key]:  # noqa: E271
            # Convert the list of accumulated arrays/lists into a single NumPy array
            elems_new[key] = np.array(elems_lst[key], dtype=int)
        elif isinstance(elems_lst[key], list) and not elems_lst[key]:
            # Determine the expected number of columns
            elems_new[key] = np.empty((0, faceNum[faceType.index(key)]), dtype=int)

    for key in csets_lst:
        csets_new[key] = tuple(np.array(lst, dtype=int) for lst in csets_lst[key])

    # Convert points_list back to a NumPy array
    points = np.array(pointl)

    mesh   = meshio.Mesh(points    = points,     # noqa: E251
                         cells     = elems_new,  # noqa: E251
                         cell_sets = csets_new)  # noqa: E251

    hopout.sep()

    return mesh


@cache
def split_hex_to_tets(order: int) -> list[tuple]:
    """
    Given the indices of a single hexahedral element, return a list of new tetra element connectivity lists

    The node numbering convention assumed here (c0, c1, c2, c3, c4, c5, c6, c7) is the usual:
          7-------6
         /|      /|
        4-------5 |
        | 3-----|-2
        |/      |/
        0-------1

    """
    # Perform the 6-tet split of the cube-like cell
    match order:
        case 1:
            # 1. strategy: 6 tets per box, all tets have same volume and angle, not periodic but isotropic
            return [( 0, 2, 3, 4),
                    ( 0, 1, 2, 4),
                    ( 2, 4, 6, 7),
                    ( 2, 4, 5, 6),
                    ( 1, 2, 4, 5),
                    ( 2, 3, 4, 7)]
            # ! 2. strategy: 6 tets per box, split hex into two prisms and each prism into 3 tets, periodic but strongly anisotropic
            # c0, c1, c2, c3, c4, c5, c6, c7 = nodes
            #  return [[0, 1, 3, 4],
            #          [1, 4, 5, 7],
            #          [1, 3, 4, 7],
            #          [1, 2, 5, 7],
            #          [1, 3, 2, 7],
            #          [2, 5, 7, 6]]
        case 2:
            return [( 0,  2,  3,  4,  24,  10,  11,  16,  26,  20),
                    ( 0,  1,  2,  4,   8,   9,  24,  16,  22,  26),
                    ( 2,  4,  6,  7,  26,  25,  18,  23,  15,  14),
                    ( 2,  4,  5,  6,  26,  12,  21,  18,  25,  13),
                    ( 1,  2,  4,  5,   9,  26,  22,  17,  21,  12),
                    ( 2,  3,  4,  7,  10,  20,  26,  23,  19,  15)]
        case 4:
            tetra1 = (  0,   2,   3,   4,  80,  88,  82,  14,  15,  16,  # noqa: E501
                       19,  18,  17,  32,  33,  34, 100, 124, 102,
                       47,  52,  45,  98, 122, 114, 108, 101, 118,
                       44,  51,  48,  84,  85,  81, 109)
            tetra2 = (  0,   1,   2,   4,   8,   9,  10,  11,  12,  13,  # noqa: E501
                       82,  88,  80,  32,  33,  34,  63,  70,  65,
                      100, 124, 102,  62,  66,  69,  99, 107, 120,
                       98, 122, 114,  87,  83,  86, 106)
            tetra3 = (  2,   4,   6,   7, 100, 124, 102,  89,  97,  91,  # noqa: E501
                       40,  39,  38,  71,  79,  73,  29,  30,  31,
                       26,  27,  28, 121, 113, 105,  96,  95,  92,
                       78,  74,  77, 116, 123, 104, 112)
            tetra4 = (  2,   4,   5,   6, 100, 124, 102,  20,  21,  22,  # noqa: E501
                       56,  61,  54,  38,  39,  40,  89,  97,  91,
                       23,  24,  25, 116, 123, 104,  93,  90,  94,
                       58,  59,  55, 119, 110, 103, 111)
            tetra5 = (  1,   2,   4,   5,  11,  12,  13, 100, 124, 102,  # noqa: E501
                       65,  70,  63,  35,  36,  37,  54,  61,  56,
                       20,  21,  22,  53,  57,  60, 119, 110, 103,
                       67,  68,  64,  99, 107, 120, 115)
            tetra6 = (  2,   3,   4,   7,  14,  15,  16,  47,  52,  45,  # noqa: E501
                      102, 124, 100,  71,  79,  73,  41,  42,  43,
                       29,  30,  31,  75,  72,  76,  50,  49,  46,
                      121, 113, 105, 108, 101, 118, 117)

            return [tetra1, tetra2, tetra3, tetra4, tetra5, tetra6]
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def tetra_faces(order: int) -> tuple[np.ndarray, ...]:
    """
    Given the tetrahedral indices, return the 4 triangular faces as tuples
    """
    match order:
        case 1:
            return (np.array((  0,  1,  2), dtype=int),
                    np.array((  0,  1,  3), dtype=int),
                    np.array((  0,  2,  3), dtype=int),
                    np.array((  1,  2,  3), dtype=int))
        case 2:
            return (np.array((  0,  1,  2,  4,  5,  6), dtype=int),
                    np.array((  0,  1,  3,  4,  8,  7), dtype=int),
                    np.array((  0,  2,  3,  6,  9,  7), dtype=int),
                    np.array((  1,  2,  3,  5,  9,  8), dtype=int))
        case 4:
            return (np.array((  0,  1,  2,  *range( 4, 13)          , *range(31, 34)), dtype=int),
                    np.array((  0,  1,  3,  *range( 4,  7)          , *range(16, 19), *reversed(range(13, 16)), *range(22, 25)), dtype=int),  # noqa: E501
                    np.array((  0,  2,  3,  *reversed(range(10, 13)), *range(19, 22), *reversed(range(13, 16)), *range(28, 31)), dtype=int),  # noqa: E501
                    np.array((  1,  2,  3,  *range( 7, 10)          , *range(19, 22), *reversed(range(16, 19)), *range(25, 28)), dtype=int))  # noqa: E501
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def split_hex_to_pyram(order: int) -> list[tuple[int, ...]]:
    """
    Given the indices of a single hexahedral element, return a list of new pyramid element connectivity lists
    """
    match order:
        case 1:
            # Perform the 6-pyramid split of the cube-like cell
            return [( 0,  1,  2,  3,  8),
                    ( 0,  4,  5,  1,  8),
                    ( 1,  5,  6,  2,  8),
                    ( 0,  3,  7,  4,  8),
                    ( 4,  7,  6,  5,  8),
                    ( 6,  7,  3,  2,  8)]
            # 3-pyramid split
            # return [( 0, 1, 2, 3, 4),
            #         ( 1, 5, 6, 2, 4),
            #         ( 6, 7, 3, 2, 4)]
        case 2:
            # Perform the 6-pyramid split of the cube-like cell
            return [(  0,  1,  2,  3, 26,  8,  9, 10, 11, 27, 28, 30, 29, 24),
                    (  0,  4,  5,  1, 26, 16, 12, 17,  8, 27, 31, 32, 28, 22),
                    (  1,  5,  6,  2, 26, 17, 13, 18,  9, 28, 32, 34, 30, 21),
                    (  0,  3,  7,  4, 26, 11, 19, 15, 16, 27, 29, 33, 31, 20),
                    (  4,  7,  6,  5, 26, 15, 14, 13, 12, 31, 33, 34, 32, 25),
                    (  6,  7,  3,  2, 26, 14, 19, 10, 18, 34, 33, 29, 30, 23)]
            # 3-pyramid split
            # return [(  0,  1,  2,  3,  4,  8,  9, 10, 11, 16, 22, 26, 20, 24),
            #         (  1,  5,  6,  2,  4, 17, 13, 18,  9, 22, 12, 25, 26, 21),
            #         (  6,  7,  3,  2,  4, 14, 19, 10, 18, 25, 15, 20, 26, 23)]
        case 4:
            return [(  0,   1,   2,   3, 124, *range( 8,  17), *reversed(range(17, 20)),
                     125,  98, 146, 128,  99, 147, 140, 100, 151, 137, 101, 150, 126,
                     127, 106, 132, 136, 107, 138, 139, 108, 129, 133, 109,  80,  83,
                      82,  81,  87,  86,  85,  84,  88, 130, 131, 135, 134, 122),
                    (  0,   4,   5,   1, 124, *range(32,  35),          *range(20,  23),
                      37,  36,  35,  10,   9,   8, 125,  98, 146, 173, 102, 162, 176,
                     103, 163, 128,  99, 147, 141, 157, 114, 174, 175, 110, 144, 160,
                     115, 126, 127, 106,  62,  65,  64,  63,  69,  68,  67,  66,  70, 142, 158, 159, 143, 120),
                    (  1,   5,   6,   2, 124, *range(35,  38),          *range(23,  26),
                      40,  39,  38,  13,  12,  11, 128,  99, 147, 176, 103, 163, 188,
                     104, 167, 140, 100, 151, 144, 160, 115, 180, 184, 111, 156, 172,
                     116, 132, 136, 107,  53,  56,  55,  54,  60,  59,  58,  57,  61, 148, 164, 168, 152, 119),
                    (  0,   3,   7,   4, 124,  17,  18,  19, *range(41,  44), *reversed(range(29, 32)),
                      34,  33,  32, 125,  98, 146, 137, 101, 150, 185, 105, 166, 173,
                     102, 162, 129, 133, 109, 153, 169, 117, 177, 181, 113, 141, 157,
                     114,  44,  47,  46,  45,  51,  50,  49,  48,  52, 145, 149, 165, 161, 118),
                    (  4,   7,   6,   5, 124, *range(29,  32), *reversed(range(26, 29)),
                      25,  24,  23,  22,  21,  20, 173, 102, 162, 185, 105, 166, 188,
                     104, 167, 176, 103, 163, 177, 181, 113, 186, 187, 112, 180, 184,
                     111, 174, 175, 110,  89,  92,  91,  90,  96,  95,  94,  93,  97, 178, 182, 183, 179, 123),
                    (  6,   7,   3,   2, 124, *range(26,  29), *reversed(range(41, 44)),
                      16,  15,  14,  38,  39,  40, 188, 104, 167, 185, 105, 166, 137,
                     101, 150, 140, 100, 151, 187, 186, 112, 169, 153, 117, 139, 138,
                     108, 172, 156, 116,  74,  73,  72,  71,  77,  76,  75,  78,  79, 171, 170, 154, 155, 121)]
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def pyram_faces(order: int) -> tuple[np.ndarray, ...]:
    """
    Given the pyramid corner indices, return the 4 triangular faces and 1 quadrilateral face as tuples
    """
    match order:
        case 1:
            return (# Triangular faces  # noqa: E261
                    np.array((  0,  1,  4), dtype=int),
                    np.array((  1,  2,  4), dtype=int),
                    np.array((  2,  3,  4), dtype=int),
                    np.array((  3,  0,  4), dtype=int),
                    # Quadrilateral face
                    np.array((  0,  1,  2,  3), dtype=int))
        case 2:
            return (# Triangular faces  # noqa: E261
                    np.array((  0,  1,  4,  5, 10,  9), dtype=int),  # 8, 22,16
                    np.array((  1,  2,  4,  6, 11, 10), dtype=int),  # 9, 26,22
                    np.array((  2,  3,  4,  7, 12, 11), dtype=int),  # 10,20,26
                    np.array((  3,  0,  4,  8,  9, 12), dtype=int),  # 11,16,20
                    # Quadrilateral face
                    np.array((  0,  1,  2,  3,  5,  6,  7,  8, 13), dtype=int))
        case 4:
            return (# Triangular faces  # noqa: E261
                    np.array((  0,  1,  4,  *range( 4,  7), *range(19, 22), *reversed(range(16, 19)), *range(28, 31)), dtype=int),  # noqa: E501
                    np.array((  1,  2,  4,  *range( 7, 10), *range(22, 25), *reversed(range(19, 22)), *range(31, 34)), dtype=int),  # noqa: E501
                    np.array((  2,  3,  4,  *range(10, 13), *range(25, 28), *reversed(range(22, 25)), *range(34, 37)), dtype=int),  # noqa: E501
                    np.array((  3,  0,  4,  *range(13, 16), *range(16, 19), *reversed(range(25, 28)), *range(37, 40)), dtype=int),  # noqa: E501
                    # Quadrilateral face
                    np.array(( 0,  1,  2,  3, *range(5, 17), *range(41, 50)), dtype=int))
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def split_hex_to_prism(order: int) -> list[tuple[int, ...]]:
    """
    Given the indices of a single hexahedral element, return a list of new prism element connectivity lists
    """
    match order:
        case 1:
            #  return [( 0,  1,  3,  4,  5,  7),
            #          ( 1,  2,  3,  5,  6,  7)]
            return [( 0,  1,  2,  4,  5,  6),
                    ( 0,  2,  3,  4,  6,  7)]
        case 2:
            #  HEXA: [ 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 24 22 21 23 20 25 26]
            #  return [(  0,  1,  3,  4,  5,  7,  8, 24, 11, 12, 25, 15, 16, 17, 19, 22, 26, 20),
            #          (  1,  2,  3,  5,  6,  7,  9, 10, 24, 13, 14, 25, 17, 18, 19, 21, 23, 26)]
            return [(  0,  1,  2,  4,  5,  6,  8,  9, 24, 12, 13, 25, 16, 17, 18, 22, 21, 26),
                    (  0,  2,  3,  4,  6,  7, 24, 10, 11, 25, 14, 15, 16, 18, 19, 26, 23, 20)]
        case 4:
            # prism1 = (   0,   1,   3,   4,   5,   7,
            #              8,   9,  10,  83,  88,  81,  19,  18,  17,            # 6 vertices
            #             20,  21,  22,  90,  97,  92,  31,  30,  29,            # Edge
            #             32,  33,  34,  35,  36,  37,  41,  42,  43,            # Edge
            #             62,  63,  64,  65,  66,  67,  68,  69,  70,            # Face 1
            #             99, 101, 105, 103, 122, 117, 123, 115, 124,            # Face 2
            #             47,  44,  45,  46,  51,  48,  49,  50,  52,            # Face 3
            #             89,  93,  96,                                          # Face 4
            #             80,  87,  84,                                          # Face 5
            #             98, 106, 109, 114, 120, 118, 102, 110, 113)            # Volume
            prism1 = (   0,   1,   3,   4,   5,   7,                           # 6 vertices
                        *range( 8, 11), 83, 88, 81, *reversed(range(17, 20)),  # Edges
                        *range(20, 23), 90, 97, 92, *reversed(range(29, 32)),  # Edges
                        *range(32, 35), 35, 36, 37, *reversed(range(41, 44)),  # Face 1
                        *range(62, 65), 65, 66, 67,          *range(68, 71) ,  # Face 2
                         99,  101, 105, 103, 122,  117, 123, 115, 124,         # Face 3
                         47, *range(44, 47),  51, *range(48, 51),  52,         # Face 4
                         89,  93,  96, 80,  87,  84,                           # Face 5
                         98, 106, 109, 114, 120, 118, 102, 110, 113)           # Volume
            # prism2 = (   1,   2,   3,   5,   6,   7,                           # 6 vertices
            #             11,  12,  13,  14,  15,  16,  81,  88,  83,            # Edges
            #             23,  24,  25,  26,  27,  28,  92,  97,  90,            # Edges
            #             35,  36,  37,  38,  39,  40,  41,  42,  43,            # Face 1
            #             53,  54,  55,  56,  57,  58,  59,  60,  61,            # Face 2
            #             71,  72,  73,  74,  75,  76,  77,  78,  79,            # Face 3
            #            101,  99, 103, 105, 122, 115, 123, 117, 124,            # Face 4
            #             94,  91,  95, 86,  82,  85,                            # Face 5
            #            107, 100, 108, 119, 116, 121, 111, 104, 112)            # Volume
            prism2 = (   1,   2,   3,   5,   6,   7,                           # 6 vertices
                        *range(11, 14), *range(14, 17), 81, 88, 83,            # Edges
                        *range(23, 26), *range(26, 29), 92, 97, 90,            # Edges
                        *range(35, 38), *range(38, 41), 41, 42, 43,            # Face 1
                        *range(53, 56), *range(56, 59), 59, 60, 61,            # Face 2
                        *range(71, 74), *range(74, 77), 77, 78, 79,            # Face 3
                        101,  99, 103, 105, 122, 115, 123, 117, 124,           # Face 4
                         94,  91,  95,  86,  82,  85,                          # Face 5
                        107, 100, 108, 119, 116, 121, 111, 104, 112)           # Volume
            return [prism1, prism2]

        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def prism_faces(order: int) -> tuple[np.ndarray, ...]:
    """
    Given the 6 prism corner indices, return the 2 triangular and 3 quadrilateral faces as tuples.
    """
    match order:
        case 1:
            return (# Triangular faces  # noqa: E261
                    np.array((  0,  1,  2    ), dtype=int),
                    np.array((  3,  4,  5    ), dtype=int),
                    # Quadrilateral faces
                    np.array((  0,  1,  4,  3), dtype=int),
                    np.array((  1,  2,  5,  4), dtype=int),
                    np.array((  2,  0,  3,  5), dtype=int))
        case 2:
            return (# Triangular faces  # noqa: E261
                    np.array((  0,  1,  2,  6,  7,  8            ), dtype=int),
                    np.array((  3,  4,  5,  9, 10, 11            ), dtype=int),
                    # Quadrilateral faces
                    np.array((  0,  1,  4,  3,  6, 13,  9, 12, 15), dtype=int),
                    np.array((  1,  2,  5,  4,  7, 14, 10, 13, 16), dtype=int),
                    np.array((  2,  0,  3,  5,  8, 12, 11, 14, 17), dtype=int))
        case 4:
            return (# Triangular faces  # noqa: E261
                    np.array((  0, 1, 2, *range( 6, 15), *range(63, 66)), dtype=int),  # z-
                    np.array((  3, 4, 5, *range(15, 24), *range(60, 63)), dtype=int),  # z+
                    # Quadrilateral faces
                    np.array((  0, 1, 4, 3, *range( 6,  9), *range(27, 30), *reversed(range(15, 18)), *reversed(range(24, 27)), *range(33, 42)), dtype=int),  # noqa: E501
                    np.array((  1, 2, 5, 4, *range( 9, 12), *range(30, 33), *reversed(range(18, 21)), *reversed(range(27, 30)), *range(42, 51)), dtype=int),  # noqa: E501
                    np.array((  2, 0, 3, 5, *range(12, 15), *range(24, 27), *reversed(range(21, 24)), *reversed(range(30, 33)), *range(51, 60)), dtype=int))  # noqa: E501
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


# Dummy function for hexahedral elements
@cache
def split_hex_to_hex(order: int) -> list[tuple[int, ...]]:
    nodes = np.arange((order + 1) ** 3)
    return [tuple(nodes)]


# Dummy function for hexahedral elements
@cache
def hex_faces(order: int) -> tuple[np.ndarray, ...]:
    """ Given the indices of a hexahedral element, return the 6 faces as tuples
    """
    match order:
        case 1:
            return (np.array((  0,  1,  2,  3), dtype=int),
                    np.array((  0,  1,  5,  4), dtype=int),
                    np.array((  1,  2,  6,  5), dtype=int),
                    np.array((  2,  6,  7,  3), dtype=int),
                    np.array((  0,  4,  7,  3), dtype=int),
                    np.array((  4,  5,  6,  7), dtype=int))
        case 2:
            return (np.array((  0,  1,  2,  3,  8,  9, 10, 11, 24), dtype=int),
                    np.array((  0,  1,  5,  4,  8, 17, 12, 16, 22), dtype=int),
                    np.array((  1,  2,  6,  5,  9, 18, 13, 17, 21), dtype=int),
                    np.array((  2,  6,  7,  3, 18, 14, 19, 10, 23), dtype=int),
                    np.array((  0,  4,  7,  3, 16, 15, 19, 11, 20), dtype=int),
                    np.array((  4,  5,  6,  7, 12, 13, 14, 15, 25), dtype=int))
        case 3:
            return (np.array((  0,  1,  2,  3, *range( 8, 10), *range(10, 12),          *range(12, 14),  *reversed(range(14, 16)), 48, *reversed(range(50, 52)), 49), dtype=int),  # noqa: E501
                    np.array((  0,  1,  5,  4, *range( 8, 10), *range(26, 28), *reversed(range(16, 18)), *reversed(range(24, 26)), 40,          *range(41, 43) , 43), dtype=int),  # noqa: E501
                    np.array((  1,  2,  6,  5, *range(10, 12), *range(28, 30), *reversed(range(18, 20)), *reversed(range(26, 28)), 36,          *range(37, 39) , 39), dtype=int),  # noqa: E501
                    np.array((  2,  6,  7,  3, *range(28, 30), *range(20, 22), *reversed(range(30, 32)), *reversed(range(12, 14)), 44, *reversed(range(46, 48)), 45), dtype=int),  # noqa: E501
                    np.array((  0,  4,  7,  3, *range(24, 26), *range(22, 24), *reversed(range(32, 34)), *reversed(range(14, 16)), 32,          *range(33, 35) , 35), dtype=int),  # noqa: E501
                    np.array((  4,  5,  6,  7, *range(16, 18), *range(18, 20),          *range(20, 22) , *reversed(range(22, 24)), 52,          *range(53, 55) , 54), dtype=int))  # noqa: E501
        case 4:
            return (np.array((  0,  1,  2,  3, *range( 8, 11), *range(11, 14),          *range(14, 17) , *reversed(range(17, 20)), 80, *reversed(range(81, 84)), 87, *reversed(range(84, 87)), 88), dtype=int),  # noqa: E501
                    np.array((  0,  1,  5,  4, *range( 8, 11), *range(35, 38), *reversed(range(20, 23)), *reversed(range(32, 35)), 62,          *range(63, 66) , 66,          *range(67, 70) , 70), dtype=int),  # noqa: E501
                    np.array((  1,  2,  6,  5, *range(11, 14), *range(38, 41), *reversed(range(23, 26)), *reversed(range(35, 38)), 53,          *range(54, 57) , 57,          *range(58, 61) , 61), dtype=int),  # noqa: E501
                    np.array((  2,  6,  7,  3, *range(38, 41), *range(26, 29), *reversed(range(41, 44)), *reversed(range(14, 17)), 71, *reversed(range(72, 75)), 78, *reversed(range(75, 78)), 79), dtype=int),  # noqa: E501
                    np.array((  0,  4,  7,  3, *range(32, 35), *range(29, 32), *reversed(range(41, 44)), *reversed(range(17, 20)), 44,          *range(45, 48) , 48,          *range(49, 52) , 52), dtype=int),  # noqa: E501
                    np.array((  4,  5,  6,  7, *range(20, 23), *range(23, 26),          *range(26, 29) , *reversed(range(29, 32)), 89,          *range(90, 93) , 93,          *range(94, 97) , 97), dtype=int))  # noqa: E501
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import math
import sys
import traceback
from collections import defaultdict
from itertools import chain
from typing import Dict, Tuple, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def copysign_int(x: int, y: int) -> int:
    """ Return a int with the magnitude (absolute value) of x but the sign of y
    """
    return int(math.copysign(x, y))


def FEMConnect() -> None:
    """ Generate connectivity information for edges and vertices
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import CountOption, GetLogical
    from pyhope.mesh.mesh_common import edges, edge_to_corner
    # ------------------------------------------------------

    if CountOption('doFEMConnect') == 0:
        return None

    hopout.separator()
    hopout.info('GENERATE FINITE ELEMENT METHOD (FEM) CONNECTIVITY...')
    hopout.sep()

    doFEMConnect = GetLogical('doFEMConnect')
    if not doFEMConnect:
        hopout.separator()
        return None

    elems     = mesh_vars.elems
    periNodes = mesh_vars.periNodes

    # Create a bidirectional lookup using a single dictionary comprehension with chain
    periDict = { k: v for k, v in chain(((int(node), int(peri)) for (node, _), peri in periNodes.items()),
                                        ((int(peri), int(node)) for (node, _), peri in periNodes.items()))}

    # Build mapping of each node -> set of element indices that include that node.
    nodeToElements = defaultdict(set)
    for idx, elem in enumerate(elems):
        for n in elem.nodes:
            nodeToElements[int(n)].add(idx)

    # Precompute combined connectivity for each node
    # > For a given node, the combined set is:
    # > nodeToElements[node] ∪ nodeToElements[periDict[node]]
    nodeConn = { node: elemSet.union(nodeToElements.get(periDict.get(node), set()))
                 for node, elemSet in nodeToElements.items()}

    # Collect all unique canonical vertices from every element
    # > The canonical vertex is the minimum of the node and its periodic counterpart
    canonicalSet = { min(int(node), periDict.get(int(node), int(node))) for elem in elems
                                                                        for node in elem.nodes[:(elem.type % 100)]}

    # Create a mapping from each canonical vertex to a unique index
    # > FEMVertexID starts at 1
    sortedCanonical = sorted(canonicalSet)
    FEMNodeMapping  = { canonical: newID for newID, canonical in enumerate(sortedCanonical, start=1)}

    # Build the vertex connectivity
    for idx, elem in enumerate(elems):
        vertexInfo: Dict[int, Tuple[int, Tuple[int, ...]]] = {}
        for locNode, node in enumerate(int(n) for n in elem.nodes[:elem.type % 100]):
            # Determine canonical vertex id
            canonical   = min(node, periDict.get(node, node))
            FEMVertexID = FEMNodeMapping[canonical]
            # Retrive connectivity set for the node
            nodeVertex = nodeConn.get(node, set())
            vertexInfo[locNode] = (FEMVertexID, tuple(sorted(nodeVertex)))
        # Set the vertex connectivity for the element
        elem.vertexInfo = vertexInfo

        # Build the edge connectivity
        # > Loop over all edges of an element
        # for iEdge, edge in enumerate(edge_to_corner):
        edgeInfo: Dict[int, Tuple[int, int | None, Tuple[int, ...], Tuple[int, ...]]] = {}
        for iEdge in edges(elem.type):
            # Get the nodes of the edge
            edge        = edge_to_corner(iEdge, elem.type)
            edge        = tuple(int(s) for s in elem.nodes[edge])
            # Determine canonical vertex ID
            canonical   = [min(edge[s], cast(int, periDict.get(edge[s], edge[s]))) for s in range(2)]
            # Get the FEM vertex ID
            FEMVertexID = tuple(FEMNodeMapping[c] for c in canonical)
            # Set the edge connectivity for the element
            # > FEMVertexID is unsorted as it contains the global orientation of the edge
            edgeInfo[iEdge] = (iEdge, None, FEMVertexID, edge)
        # Set the edge information for the element
        elem.edgeInfo = edgeInfo


def getFEMInfo(nodeInfo: np.ndarray) -> tuple[np.ndarray,  # FEMElemInfo
                                              np.ndarray,  # VertexInfo
                                              np.ndarray,  # VertexConnectInfo
                                              np.ndarray,  # EdgeInfo
                                              np.ndarray   # EdgeConnectInfo
                                             ]:
    """ Extract the FEM connectivity information and return four arrays

     - FEMElemInfo      : [offsetIndEdge, lastIndEdge, offsetIndVertex, lastIndVertex]
     - vertexInfo       : [FEMVertexID, offsetIndVertexConnect, lastIndVertexConnect]
     - vertexConnectInfo: [nbElemId, nbLocVertexId]
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    elems  = mesh_vars.elems
    nElems = len(elems)

    # Check if elements contain FEM connectivity
    if not hasattr(elems[0], 'vertexInfo') or elems[0].vertexInfo is None:
        return np.array([]), np.array([]), np.array([]), np.array([]), np.array([])

    # Vertex connectivity info ---------------------------------------------------
    # > Build list of all vertex occurrences, appearing in the same order as the elements
    occList = [(FEMVertexID, elemID, locNode) for elemID , elem             in enumerate(elems)  # noqa: E272
                                              for locNode, (FEMVertexID, _) in elem.vertexInfo.items()]

    # > Build mapping from FEM vertex ID to list of occurrences
    groups = defaultdict(list)
    for occIdx, (FEMVertexID, elemID, locNode) in enumerate(occList):
        groups[FEMVertexID].append((occIdx, elemID, locNode))

    # Initialize FEM element information
    FEMElemInfo    = np.zeros((nElems, 4), dtype=np.int32)

    vertexInfoList = []  # List: [FEMVertexID, offsetIndVertexConnect, lastIndVertexConnect]
    vertexConnList = []  # List: [[nbElemID, nbLocVertexID]]
    vertexOffset   = 0
    occGlobalIdx   = 0   # global index in occList

    for elemID, elem in enumerate(elems):
        # Process vertex occurrences for the current element
        for _ in elem.vertexInfo:
            # Get the occurrence information from the global occList
            FEMVertexID, _, locNode = occList[occGlobalIdx]
            groupOcc = groups[FEMVertexID]
            offset   = len(vertexConnList)

            # Identify the master occurrence (lowest occIdx from the occurrence group)
            masterOcc = min(x[0] for x in groupOcc)

            # Build connectivity list for current element, excluding itself
            connections = [(nbElem+1, nbLocal+1) if   otherOcc == masterOcc else (-(nbElem+1), nbLocal+1)  # noqa: E271
                                                 for (otherOcc, nbElem, nbLocal) in groupOcc if otherOcc != occGlobalIdx]

            if connections:
                lastIndex = offset + len(connections)
                vertexConnList.extend(connections)
            else:  # No connections
                lastIndex = offset

            # Append vertex information
            vertexInfoList.append([FEMVertexID, offset, lastIndex])
            occGlobalIdx += 1

        # Set the vertex connectivity offset for this element.
        FEMElemInfo[elemID, 2] = vertexOffset
        FEMElemInfo[elemID, 3] = vertexOffset + len(elem.vertexInfo)
        vertexOffset += len(elem.vertexInfo)

    # Edge   connectivity info ---------------------------------------------------
    edgeInfoList   = []  # List: [FEMEdgeID  , offsetIndEdgeConnect  , lastIndEdgeConnect]
    edgeConnList   = []  # List: [[nbElemID, nbLocEdgeID]]
    edgGlobalIdx   = 0   # global edge index

    for elemID, elem in enumerate(elems):
        # Process edge occurrences for the current element
        for iEdge, (locEdge, edgeIdx, edge, edgeNodes) in enumerate(elem.edgeInfo.values()):
            # Get the elements connected to both edge nodes
            groupOcc  = [groups[e] for e in edge]
            offset    = len(edgeConnList)

            # Identify elements connected to the current edge
            edgeElems = set([e[1]  for g in groupOcc for e in g])  # noqa: E272

            # Build connectivity list for current element, excluding the current edge
            connections = [(nbElem, nbEdgeIdx, nbLocEdge, nbEdge, nbEdgeNodes)
                           for nbElem                                      in edgeElems                            # noqa: E272
                           for (nbLocEdge, nbEdgeIdx, nbEdge, nbEdgeNodes) in elems[nbElem].edgeInfo.values()      # noqa: E272
                           if set(nbEdge) == set(edge) and (nbLocEdge != locEdge or nbElem != elem.elemID)]

            if connections:
                # Sanity check, all edge should have the same global index
                if ((edgeIdx is     None and any(     e  is not None for (_, e, _, _, _) in connections)) or       # noqa: E271, E272
                    (edgeIdx is not None and len({abs(e)             for (_, e, _, _, _) in connections}) != 1)):  # noqa: E271, E272
                    hopout.warning('FEMConnect: Inconsistent edge global index')
                    traceback.print_stack(file=sys.stdout)
                    sys.exit(1)

            # Check if the edges already have an global index
            if edgeIdx is not None:
                # Check if the current edge is the master edge
                if edgeIdx > 0:  # master edge
                    masterID, masterEdge, masterEdgeNodes = -1, edge, edgeNodes
                else:            # slave edge
                    # Find the master edge among the connections
                    masterID = [i for i in range(len(connections)) if connections[i][1] > 0]
                    if len(masterID) != 1:
                        hopout.warning('FEMConnect: Inconsistent edge global index')
                        traceback.print_stack(file=sys.stdout)
                        sys.exit(1)
                    masterID = masterID[0]
                    masterID, masterEdge, masterEdgeNodes = masterID, *connections[masterID][3:5]
            # Otherwise, the current edge is the master edge and the others are slave edges
            else:
                edgGlobalIdx   +=  1
                edgeIdx         = edgGlobalIdx
                masterID, masterEdge, masterEdgeNodes = -1, edge, edgeNodes
                # Set the current edge as master edge
                elem.edgeInfo[iEdge] = (locEdge, edgGlobalIdx, edge, edgeNodes)
                # Set the global index for the other edges
                for nbElem, _, nbLocEdge, _, _ in connections:
                    e = list(elems[nbElem].edgeInfo[nbLocEdge])
                    e[1] = -edgGlobalIdx
                    elems[nbElem].edgeInfo[nbLocEdge] = tuple(e)

            # TODO: Check if the orientation of the master edge is with ascending nodeInfo index
            orientation =  1 if nodeInfo[masterEdgeNodes[0]] < nodeInfo[masterEdgeNodes[1]] else -1
            # Current edge is a slave edge, check our relative orientation
            if masterID != -1:
                # Check if the edge is oriented in the same direction
                orientation = orientation if masterEdge[0] == edge[0] else -orientation

            edgeConn = []
            for iConn, (nbElem, _, nbLocEdge, nbEdge, _) in enumerate(connections):
                # The current edge is the master
                if masterID == -1:
                    orientedElemID  = -(nbElem   +1)
                    orientedLocEdge =   nbLocEdge+1 if nbEdge   == masterEdge               else -(nbLocEdge+1)  # noqa: E272

                # The master edge is one of the connections, indicated by masterID
                else:
                    orientedElemID  =   nbElem   +1 if masterID == iConn                    else -(nbElem   +1)  # noqa: E272
                    orientedLocEdge =   nbLocEdge+1 if nbEdge   == connections[masterID][3] else -(nbLocEdge+1)

                # Append the edge connectivity
                edgeConn.append([orientedElemID, orientedLocEdge])

            if connections:
                lastIndex = offset + len(edgeConn)
                edgeConnList.extend(edgeConn)
            else:  # No connections
                lastIndex = offset

            # Append edge information
            edgeInfoList.append([copysign_int(edgeIdx, orientation), offset, lastIndex])

    # INFO: Same output as above but looping over the occurrences
    # for occIdx, (FEMVertexId, elemID, locNode) in enumerate(occList):
    #     groupOcc    = groups[FEMVertexId]
    #     offset      = len(vertexConnList)
    #
    #     # Identify the master occurrence (lowest occurrence index in the group)
    #     masterOcc   = min(x[0] for x in groupOcc)
    #
    #     # Build connectivity list for current element, excluding itself
    #     connections = [(nbElem+1, nbLocal+1) if otherOcc == masterOcc else (-(nbElem+1), nbLocal+1)
    #                    for (otherOcc, nbElem, nbLocal) in groupOcc if otherOcc != occIdx]
    #     if connections:
    #         lastIndex = offset + len(connections)
    #         vertexConnList.extend(connections)
    #     else:  # No connections
    #         lastIndex = offset
    #     # Append vertex information
    #     vertexInfoList.append([FEMVertexId, offset, lastIndex])
    #
    #     # Update FEMElemInfo for the corresponding element.
    #     # Set the offset to the minimum and last index to the maximum among occurrences.
    #     if offset < FEMElemInfo[elemID, 2]:
    #         FEMElemInfo[elemID, 2] = offset
    #     if lastIndex > FEMElemInfo[elemID, 3]:
    #         FEMElemInfo[elemID, 3] = lastIndex

    # Convert lists to numpy arrays
    vertexInfo = np.array(vertexInfoList, dtype=np.int32)
    vertexConn = np.array(vertexConnList, dtype=np.int32) if vertexConnList else np.array((0, 2), dtype=np.int32)

    edgeInfo   = np.array(edgeInfoList  , dtype=np.int32)
    edgeConn   = np.array(edgeConnList  , dtype=np.int32) if edgeConnList   else np.array((0, 2), dtype=np.int32)  # noqa: E272

    return FEMElemInfo, vertexInfo, vertexConn, edgeInfo, edgeConn



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
import os
import subprocess
import sys
import traceback
from typing import Optional, Union, cast, final
from typing_extensions import override
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import h5py
import numpy as np
from collections import OrderedDict
from configparser import ConfigParser
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


@final
class MultiOrderedDict(OrderedDict):
    """ Add option to repeat the same key multiple times

        Standard ConfigParser only supports one value per key,
        thus overload the ConfigParser with this new dict_type
    """
    @override
    def __setitem__(self, key, value) -> None:
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            super().__setitem__(key, value)


def strtobool(val: Union[int, bool, str]) -> bool:  # From distutils.util.strtobool() [Python 3.11.2]
    """ Convert a string representation of truth to True or False.
        True values  are 'y', 'yes', 't', 'true', 'on', and '1';
        False values are 'n', 'no' , 'f', 'false', 'off', and '0'.
        Raises ValueError if 'val' is anything else.
    """
    if type(val) is bool:
        return val
    if type(val) is int:
        val = str(val)
    if type(val) is str:
        val = val.lower()

    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError('invalid truth value %r' % (val,))


# ==================================================================================================================================
@final
class DefineConfig:
    """ Provide routines to define all HOPR parameters
    """
    def __init__(self) -> None:
        # Create an empty config dictionary
        self.dict = dict()
        return None

    def __enter__(self) -> dict:
        return self.dict

    def __exit__(self, *args: object) -> None:
        return None


# ==================================================================================================================================
def CheckDefined(name: str, multiple: bool = False, init: bool = False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    # Check if already defined which is allowed if
    # - we are not in init
    # - multiple parameter
    if init:
        if name in config.prms and not multiple:
            hopout.warning('Parameter "{}" already define and not a multiple option, exiting...'.format(name))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)
    else:
        if name not in config.prms:
            hopout.warning('Parameter "{}" is not defined, exiting...'.format(name))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


def CheckUsed(name: str) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if config.prms[name]['counter'] > 1 and not config.prms[name]['multiple']:
        hopout.warning('Parameter "{}" already used and not a multiple option, exiting...'.format(name))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)


def CheckType(name: str, calltype: str) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if config.prms[name]['type'] is not calltype:
        hopout.warning('Call type of parameter "{}" does not match definition, exiting...'.format(name))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)


def CheckDimension(name: str, result: int) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if config.prms[name]['number'] != result:
        hopout.warning('Parameter "{}" has array length mismatch, exiting...'.format(name))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)


def CreateSection(string: str) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple=False, init=True)
    config.prms[string] = dict(type='section', name=string)


def CreateStr(string: str, help: Optional[str] = None, default: Optional[str] = None, multiple: bool = False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='str',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateReal(string: str, help: Optional[str] = None, default: Optional[float] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='real',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateInt(string: str, help: Optional[str] = None, default: Optional[int] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='int',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateLogical(string: str, help: Optional[str] = None, default: Optional[bool] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='bool',
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateIntFromString(string: str, help: Optional[str] = None, default: Optional[str] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='int2str',
                               name=string,
                               mapping=dict(),
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateIntOption(string: str, name, number) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    multiple = config.prms[string]['multiple']
    CheckDefined(string, multiple=multiple, init=False)
    config.prms[string]['mapping'].update({number: name})


def CreateRealArray(string: str, nReals, help: Optional[str] = None, default: Optional[str] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='realarray',
                               number=nReals,
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


def CreateIntArray(string: str, nInts, help: Optional[str] = None, default: Optional[str] = None, multiple=False) -> None:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    # ------------------------------------------------------

    CheckDefined(string, multiple, init=True)
    config.prms[string] = dict(type='intarray',
                               number=nInts,
                               name=string,
                               help=help,
                               default=default,
                               counter=0,
                               multiple=multiple)


# ==================================================================================================================================
def CountOption(string: str) -> int:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    from configparser import NoOptionError
    # ------------------------------------------------------

    CheckDefined(string)

    try:
        counter = len([s for s in config.params.get('general', string).split('\n') if s != ''])
    except NoOptionError:
        counter = 0
    return counter


def GetParam(name: str, calltype: str, default: Optional[str] = None, number: Optional[int] = None):
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    CheckDefined(name)
    config.prms[name]['counter'] += 1
    CheckUsed(name)
    CheckType(name, calltype)

    if config.params.has_option('general', name):
        if config.prms[name]['multiple']:
            # We can request specific indices
            if number is None: num = config.prms[name]['counter']-1  # noqa: E701
            else:              num = number                          # noqa: E701

            input = [s for s in config.params.get('general', name).split('\n') if s != '']
            if num >= len(input):
                hopout.warning(f'Index {num+1} is out of range for option "{name}"')
                # traceback.print_stack(file=sys.stdout)
                sys.exit(1)
            value = input[num]
        else:
            value = config.params.get('general', name)
            # Single values cannot contain spaces
            if '\n' in value:
                hopout.warning(f'Option "{name}" is already set, but is not a multiple option!')
                # traceback.print_stack(file=sys.stdout)
                sys.exit(1)

        # int2str has custom output
        if calltype != 'int2str':
            if calltype == 'bool':
                hopout.printoption(name, '{0:}'.format(value), '*CUSTOM')
            else:
                hopout.printoption(name, value               , '*CUSTOM')
    else:
        if default:
            value = default
        else:
            if config.prms[name]['default'] is not None:
                value = config.prms[name]['default']

                # int2str has custom output
                if calltype != 'int2str':
                    if calltype == 'bool':
                        hopout.printoption(name, '{0:}'.format(value), 'DEFAULT')
                    else:
                        hopout.printoption(name, value               , 'DEFAULT')
            else:
                hopout.warning(f'Keyword "{name}" not found in file and no default given, exiting...')
                # traceback.print_stack(file=sys.stdout)
                sys.exit(1)
    return value


def GetStr(name: str, default: Optional[str] = None, number: Optional[int] = None) -> str:
    value = GetParam(name=name, default=default, number=number, calltype='str')
    return value


def GetReal(name: str, default: Optional[str] = None, number: Optional[int] = None) -> float:
    value = GetParam(name=name, default=default, number=number, calltype='real')
    return float(value)


def GetInt(name: str, default: Optional[str] = None, number: Optional[int] = None) -> int:
    value = GetParam(name=name, default=default, number=number, calltype='int')
    return int(value)


def GetLogical(name: str, default: Optional[str] = None, number: Optional[int] = None) -> bool:
    value = GetParam(name=name, default=default, number=number, calltype='bool')
    return strtobool(value)


def GetIntFromStr(name: str, default: Optional[str] = None, number: Optional[int] = None) -> int:
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    value  = GetParam(name=name, default=default, number=number, calltype='int2str')
    source = 'DEFAULT' if config.prms[name]['counter'] == 0 else '*CUSTOM'

    # Check if we already received the int. Otherwise, get the value from the mapping
    mapping = config.prms[name]['mapping']
    if type(value) is int:
        hopout.printoption(name, '{} [{}]'.format(value, mapping[value]), source)
    else:
        if not value.isdigit():
            value = [s for s, v in mapping.items() if v.lower() == value.lower()]
            if len(value) == 0:
                hopout.warning('Unknown value for parameter {}, exiting...'.format(name))
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)
            else:
                value = int(value[0])
                hopout.printoption(name, '{} [{}]'.format(value, mapping[value]), source)
        else:
            hopout.printoption(name, '{} [{}]'.format(value, mapping[int(value)]), source)

    return int(value)


def GetRealArray(name: str, default: Optional[str] = None, number: Optional[int] = None) -> np.ndarray:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    value = GetParam(name=name, default=default, number=number, calltype='realarray')

    # Split the array definitiosn
    value = value.split('(/')[1]
    value = value.split('/)')[0]

    # Commas separate 1st dimension, double commas separate 2nd dimension
    if ',,' in value:
        value = [s.split(',') for s in value.split(',,')]
        try:
            value = np.array(value).astype(float)
        except ValueError as e:
            print()
            print(hopout.warn(f'{e}'))
            hopout.warning(f'Failed to read "{name}" array, possibly malformed comma-separated data. Exiting...')
            sys.exit(1)
    else:
        value = value.split(',')
        try:
            value = np.array(value).astype(float)
        except ValueError as e:
            print()
            print(hopout.warn(f'{e}'))
            hopout.warning(f'Failed to read "{name}" array, possibly malformed comma-separated data. Exiting...')
            sys.exit(1)
    CheckDimension(name, value.size)
    return value


def GetIntArray(name: str, default: Optional[str] = None, number: Optional[int] = None) -> np.ndarray:
    value = GetParam(name=name, default=default, number=number, calltype='intarray')

    # Split the array definitiosn
    value = value.split('(/')[1]
    value = value.split('/)')[0]

    # Commas separate 1st dimension, double commas separate 2nd dimension
    value = [s.split(',') for s in value.split(',,')]
    value = np.array(value).astype(int)
    # Reduce dimensions
    value = np.concatenate(value).ravel()
    CheckDimension(name, value.size)
    return value


# ==================================================================================================================================
@final
class ReadConfig():
    """ Read an HOPR parameter file

        This file is meant to remain compatible to the HOPR parameter file
        format, so we need some hacks around the INI file format
    """

    def __init__(self, input: str) -> None:
        self.input     = input
        self.parameter = ''
        self.mesh      = ''

        # define allowed comments
        self.sym_comm = ('#', ';', '!')
        return None

    def _read_file(self) -> list:
        """ Read the parameter file and replace DEFVAR variables
        """
        # Local imports ----------------------------------------
        import pyhope.output.output as hopout
        # ------------------------------------------------------
        processed_lines = []
        variables       = {}

        with open(self.parameter, 'r', encoding='utf-8') as stream:
            for line in stream:
                # Remove all whitespaces
                line = ''.join(line.split())

                # HOPR supported inline comments as prefix before '%'
                # For legacy reasons also support such comment constructs
                if '%' in line:
                    line = line.split('%', 1)[1].strip()

                # Split of [#, ;, !] comments
                for symbol in self.sym_comm:
                    if symbol in line:
                        line = line.split(symbol, 1)[0].strip()

                # HOPR supported inline variable definitions with prefix 'DEFVAR='
                # For legacy reasons also support such variable definition constructs
                if line.strip().startswith('DEFVAR='):
                    if ':' not in line:
                        hopout.warning('DEFVAR= syntax error while parsing parameter file. Missing ":"')
                        sys.exit(1)
                    parts = line.split(':')

                    var_type_part = parts[0].replace('DEFVAR=', '').strip()
                    var_def_part  = parts[1].strip()

                    # Check if comment is in value part
                    for symbol in self.sym_comm:
                        if symbol in var_def_part:
                            var_def_part = var_def_part.split(symbol, 1)[0].strip()  # Take the part before the symbol
                            break  # Stop at the first symbol found

                    # Extract variable type and optional array size
                    if '~' in var_type_part:
                        # Vector
                        _, size_part = var_type_part.split('~')
                        arr_size = int(size_part.strip(')'))  # Convert size to int
                    else:
                        # Scalar
                        arr_size = None

                    # Extract variable name and value (handling spaces around `=`)
                    if '=' in var_def_part:
                        var_name, var_value = [s.strip() for s in var_def_part.split('=', 1)]
                    else:
                        hopout.warning(f'DEFVAR= syntax error while parsing "{var_def_part}"')
                        sys.exit(1)

                    # Ensure unique variable names
                    for existing_var in variables:
                        if var_name in set(existing_var):
                            hopout.warning(f'Variable "{var_name}" is ambiguous')
                            sys.exit(1)

                    # Convert values to proper types
                    if arr_size:  # Handle array
                        values = [float(v) if '.' in v else int(v) for v in var_value.split(',')]
                        if len(values) != arr_size:
                            hopout.warning(f'Expected {arr_size} values for array "{var_name}", got {len(values)}')
                            sys.exit(1)
                        variables[var_name] = values
                    else:  # Single value
                        variables[var_name] = float(var_value) if '.' in var_value else int(var_value)

                    # We have to sort the variables according to the length of the keys in order to avoid
                    # substring replacement in the parameter file. This way it can be assured that long strings
                    # get replaced first.
                    # variables = sorted(variables.items(), key=lambda item: len(item[0]), reverse=True)
                    variables = dict(sorted(variables.items(), key=lambda item: len(item[0]), reverse=True))

                    continue  # Skip adding this line to config

                # Replace variables in the parameter file
                for var, value in variables.items():
                    if isinstance(value, list):  # Convert arrays to string format
                        replacement = f'(/{",".join(map(str, value))}/)'
                    else:
                        replacement = str(value)

                    # Ensure exact match replacement (avoiding substring issues)
                    if '=' in line:
                        tmp = line.split('=')
                        if var in tmp[1]:
                            tmp[1] = tmp[1].replace(var, replacement)
                        line = '='.join(tmp)

                processed_lines.append(line)

        return processed_lines

    def __enter__(self) -> ConfigParser:
        # Local imports ----------------------------------------
        from pyhope.common.common_vars import Common
        import pyhope.config.config as config
        import pyhope.output.output as hopout
        # ------------------------------------------------------

        parser = ConfigParser(strict=False,
                              comment_prefixes=self.sym_comm,
                              inline_comment_prefixes=self.sym_comm,
                              dict_type=MultiOrderedDict
                              )

        # Check if the file exists in argv
        if not self.input:
            process = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], shell=False, stdout=subprocess.PIPE,
                                                                                             stderr=subprocess.DEVNULL)
            common  = Common()
            program = common.program
            version = common.version
            commit  = process.communicate()[0].strip().decode('ascii')

            hopout.header(program, version, commit)
            hopout.warning('No parameter or mesh file given')
            sys.exit(1)

        # Check if file exists on drive
        if not os.path.isfile(self.input):
            process = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'], shell=False, stdout=subprocess.PIPE)
            common  = Common()
            program = common.program
            version = common.version
            commit  = process.communicate()[0].strip().decode('ascii')

            hopout.header(program, version, commit)
            hopout.warning('Parameter or mesh file [󰇘]/{} does not exist'.format(os.path.basename(self.input)))
            sys.exit(1)

        # Check if input is mesh or parameter file
        parameter_mode = False
        mesh_mode      = False

        # Check whether given input is a valid mesh or a parameter file
        if h5py.is_hdf5(self.input):
            mesh_mode = True
        else:
            try:
                with open(self.input, 'r', encoding='utf-8') as f:
                    f.read()
                parameter_mode = True
            except UnicodeDecodeError:
                hopout.warning('Parameter or mesh file [󰇘]/{} are of unknown type'.format(os.path.basename(self.input)))
                sys.exit(1)

        # Handle parameter data
        if parameter_mode:
            self.parameter = self.input

            # Sore full path of the parameter file
            config.prmfile = os.path.abspath(self.parameter)

            # HOPR does not use conventional sections, so prepend a fake section header
            parser.read_string('[general]\n' + '\n'.join(self._read_file()))

        # Handle mesh data
        if mesh_mode:
            # Set the prmfile to an empty string as it is required for searching for the mesh later in script.
            # As the mesh is however explicitly givenb, this is not required in mesh_mode
            config.prmfile = ''

            # In this mode we need to create a dummy parameter file for processing the mesh.
            # For basic processing as calculating the jacobians and performing the checks only
            # little parameters have to be defined and extracted from the file
            mesh_params = [
                '[general]',
                f'ProjectName = {os.path.splitext(os.path.basename(self.input))[0]}',
                f'Filename    = {os.path.abspath(self.input)}',
                'OutputFormat = HDF5',
                'DebugVisu    = F',
                'Mode         = external',
            ]

            # Get geometric order and boundary conditions
            with h5py.File(self.input, 'r') as f:
                # Here we use item for legacy reasons as HOPR stores scalars as arrays with one element
                NGeo    = cast(int, f.attrs['Ngeo'].item())
                BCNames = [s.decode('utf-8').strip() for s in cast(h5py.Dataset, f['BCNames'])[:]]
                BCType  = cast(h5py.Dataset, f['BCType'])[:]

            # Write geometric order info to file
            mesh_params.append(f'NGeo = {NGeo}')
            mesh_params.append(f'MeshIsAlreadyCurved = {"T" if NGeo > 1 else "F"}')

            # Setup boundary conditions
            for iBC, BC in enumerate(BCNames):
                mesh_params.append(f'BoundaryName = {BC}')
                mesh_params.append(f'BoundaryType = (/{", ".join(map(str, BCType[iBC]))}/)')

            # Join lines into a single string
            mesh_param = '\n'.join(mesh_params)

            # Parse dummy parameters
            parser.read_string(mesh_param)

        # Parse configation file either from read in parameter file or
        # recovered from a given mesh file
        config.std_length = max(len(s) for s in config.prms.keys())
        config.std_length = max(32, config.std_length+1)

        # Loop over all objects and check if they are provided
        # for key, value in config.prms.items():
        #     if value['type'] == 'section':
        #         hopout.separator()
        #         hopout.info(key)
        #         hopout.separator()
        #         continue
        #
        #     # Check if the key is given in the parameter file
        #     if parser.has_option('general', key):
        #         # Check if the value can be converted
        #         match value['type']:
        #             case 'int':
        #                 try:
        #                     str_int = int(parser.get('general', key))
        #                 except ValueError:
        #                     hopout.warning('Keywords {} cannot be converted to integer'.format(key))
        #
        #         hopout.printoption(key, parser.get('general', key),
        #                            '*CUSTOM', std_length)
        #     # Check if a default option is given
        #     else:
        #         if value['default']:
        #             hopout.printoption(key, value['default'],
        #                                'DEFAULT', std_length)
        #         else:
        #             hopout.warning('Keyword "{}" not found in file, exiting...'
        #                            .format(key))

        return parser

    def __exit__(self, *args: object) -> None:
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import importlib.metadata
import pathlib
import re
from functools import cache
from typing import Callable, Final, final
from typing_extensions import Self
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from packaging.version import Version
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


np_mtp : int  # Number of threads for multiprocessing


# PEP 318 – Decorators for Functions and Methods
# > https://peps.python.org/pep-0318/
def singleton(cls) -> Callable:
    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance


@singleton
class Common():
    def __init__(self: Self) -> None:
        self._version: Final      = self.__version__
        self._program: Final[str] = self.__program__

    @property
    @cache
    def __version__(self) -> Version:
        # Retrieve version from package metadata
        try:
            package = pathlib.Path(__file__).parent.parent.name
            version = importlib.metadata.version(package)
        # Fallback to pyproject.toml
        except importlib.metadata.PackageNotFoundError:
            pyproject = pathlib.Path(__file__).parent.parent.parent / 'pyproject.toml'
            if not pyproject.exists():
                raise FileNotFoundError(f'pyproject.toml not found at {pyproject}')

            with pyproject.open('r') as p:
                match = re.search(r'version\s*=\s*["\'](.+?)["\']', p.read())
            if not match:
                raise ValueError('Version not found in pyproject.toml')
            version = match.group(1)

        return Version(version)

    @property
    def __program__(self) -> str:
        return 'PyHOPE'

    @property
    def program(self) -> str:
        return str(self._program)

    @property
    def version(self) -> str:
        return str(self._version)


@final
class Gitlab():
    # Gitlab "python-gmsh" access
    LIB_GITLAB:  str = 'gitlab.iag.uni-stuttgart.de'
    # LIB_PROJECT  = 'libs/python-gmsh'
    LIB_PROJECT: str = '797'
    LIB_VERSION: str = '4.13.1.post1'
    LIB_SUPPORT: dict[str, dict[str, str]] = {
        "linux": {
            "x86_64" : "6d96f1f0a3a3d96f942a6cb62deaf959380d94bbe329d61fdcf3b1dd9762981d",
            "aarch64": "104fe49eeb75ee91cb237acd251533aae98fb48c7e4e16517be6c0f4ccf677da"
        },
        "darwin": {
            "arm64"  : "cf91a48a6207c3eae9321a3c97df105320a8c3777b6b5d7411ca7343ebddf187"
        }
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
import sys
from io import TextIOWrapper
from typing import cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def DefineCommon() -> None:
    """ Define general options for the entire program
    """
    # Local imports ----------------------------------------
    from pyhope.readintools.readintools import CreateInt, CreateSection
    # ------------------------------------------------------

    # Check the number of available threads
    try:
        np_aff = len(os.sched_getaffinity(0))
    except AttributeError:
        np_aff = os.cpu_count() or 1
    # Reserve two threads for the operating system and the main thread
    np_aff -= 2

    CreateSection('Common')
    CreateInt(      'nThreads',        default=np_aff,     help='Number of threads for multiprocessing')


def InitCommon() -> None:
    """ Readin general option for the entire program
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.common.common_vars as common_vars
    from pyhope.gmsh.gmsh_install import PkgsCheckGmsh
    from pyhope.readintools.readintools import GetInt
    # ------------------------------------------------------

    hopout.info('INIT PROGRAM...')

    # Check the number of available threads
    np_req = GetInt('nThreads')
    match np_req:
        case -1 | 0:  # All available cores / no multiprocessing
            np_mtp = np_req
        case _:       # Check if the number of requested processes can be provided
            # os.affinity is Linux only
            try:
                np_aff = len(os.sched_getaffinity(0))
            except AttributeError:
                np_aff = os.cpu_count() or 1
            np_mtp = min(np_req, np_aff)

    # If running under debugger, multiprocessing is not available
    if DebugEnabled():
        print(hopout.warn('Debugger detected, disabling multiprocessing!'))
        np_mtp = 0

    # Actually overwrite the global value
    common_vars.np_mtp = np_mtp

    # Check if we are using the NRG Gmsh version and install it if not
    PkgsCheckGmsh()

    # hopout.info('INIT PROGRAM DONE!')


def DebugEnabled() -> bool:
    """ Check if program runs with debugger attached
        > https://stackoverflow.com/a/77627075/23851165
    """
    try:
        if sys.gettrace() is not None:
            return True
    except AttributeError:
        pass

    try:
        if sys.monitoring.get_tool(sys.monitoring.DEBUGGER_ID) is not None:
            return True
    except AttributeError:
        pass

    return False


def IsInteractive() -> bool:
    return cast(TextIOWrapper, sys.__stdin__).isatty()


def IsDisplay() -> bool:
    """ Check if the program is running in a display environment
    """
    # Check if running on Linux, otherwise assume a display
    if not sys.platform.startswith('linux'):
        return True

    # Check for environment variables that indicate a graphical display
    return bool(os.environ.get('DISPLAY') or os.environ.get('WAYLAND_DISPLAY'))


# > https://stackoverflow.com/a/5419576/23851165
# def object_meth(object) -> list:
#     methods = [method_name for method_name in dir(object)
#                if '__' not in method_name]
#     return methods


def find_key(dict: dict[int, str], item) -> int | None:
    """ Find the first occurrence of a key in dictionary
    """
    if type(item) is np.ndarray:
        for key, val in dict.items():
            if np.all(val == item):
                return key
    else:
        for key, val in dict.items():
            if        val == item :  # noqa: E271
                return key
    return None


def find_keys(dict: dict[int, str], item) -> tuple[int, ...] | None:
    """ Find all occurrence of a key in dictionary
    """
    if type(item) is np.ndarray:
        keys = tuple(key for key, val in dict.items() if np.all(val == item))
        if len(keys) > 0:
            return keys
    else:
        keys = tuple(key for key, val in dict.items() if        val == item )  # noqa: E271
        if len(keys) > 0:
            return keys
    return None


# def find_value(dict, item):
#     """ Find key by value in dictionary
#     """
#     return dict.keys()[dict.values().index(item)]


def find_index(seq, item) -> int:
    """ Find the first occurrences of a key in a list
    """
    if type(seq) is np.ndarray:
        seq = seq.tolist()

    if type(item) is np.ndarray:
        for index, val in enumerate(seq):
            if np.all(val == item):
                return index
    else:
        for index, val in enumerate(seq):
            if        val == item :  # noqa: E271
                return index
    return -1


def find_indices(seq, item) -> tuple[int, ...]:
    """ Find all occurrences of a key in a list
    """
    if type(seq) is np.ndarray:
        seq = seq.tolist()

    start_at = -1
    locs = []
    while True:
        try:
            loc = seq.index(item, start_at+1)
        except ValueError:
            break
        else:
            locs.append(loc)
            start_at = loc
    return tuple(locs)


def lines_that_equal(     string: str, fp: list, start_idx=0) -> list[int]:
    """ Find all occurrences of a string in a file-like object
    """
    return [num for num, line in enumerate(fp[start_idx:]) if line.strip() == string]


def lines_that_contain(   string: str, fp: list, start_idx=0) -> list[int]:
    """ Find all occurrences of a string in a file-like object
    """
    return [num for num, line in enumerate(fp[start_idx:], start=start_idx) if string in line]


def lines_that_start_with(string: str, fp: list, start_idx=0) -> list[int]:
    """ Find all occurrences of a string at the start of a line in a file-like object
    """
    return [num for num, line in enumerate(fp[start_idx:]) if line.startswith(string)]


def lines_that_end_with(  string: str, fp: list, start_idx=0) -> list[int]:
    """ Find all occurrences of a string at the end of a line in a file-like object
    """
    return [num for num, line in enumerate(fp[start_idx:]) if line.rstrip().endswith(string)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import importlib.util
import os
import sys
import traceback
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def CalcStretching(nZones: int, zone: int, nElems: np.ndarray, lEdges: np.ndarray) -> np.ndarray:
    """ Calculate the stretching parameter for meshing the current zone
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import CountOption, GetRealArray
    # ------------------------------------------------------

    # Check if mesh is scaled
    nl0     = CountOption('l0')
    nFactor = CountOption('Factor')

    conditions = {(0     , 0     ): 'constant',     # Non-stretched element arrangement
                  (0     , nZones): 'factor',       # Stretched element arrangement based on factor
                  (nZones, 0     ): 'ratio',        # Stretched element arrangement based on ratio
                  (nZones, nZones): 'combination'   # Stretched element arrangement with a combination of l0 and factor
                 }

    stretchingType = conditions.get((nl0, nFactor), None)

    if stretchingType == 'combination':
        print(hopout.warn('Both l0 and a stretching factor are provided. ' +
                          'The number of elements will be adapted to account for both parameters.'))
    if stretchingType is None:
        hopout.warning('Streching parameters not defined properly. Check whether l0 and/or Factor are defined nZone-times.')
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)

    # Calculate the stretching parameter for meshing the current zone
    stretchingHandlers = {
            'constant':    lambda: ([1., 1., 1.], None, None),                         # noqa: E272
            'factor':      lambda: (GetRealArray('Factor', number=zone), None, None),  # noqa: E272
            'ratio':       lambda: (None,                                              # noqa: E272
                                    GetRealArray('l0'    , number=zone),
                                    np.divide(lEdges, np.abs(GetRealArray('l0', number=zone)),
                                              out=np.zeros_like(lEdges), where=GetRealArray('l0', number=zone) != 0)),
            'combination': lambda: (GetRealArray('Factor', number=zone),
                                    GetRealArray('l0'    , number=zone),
                                    np.divide(lEdges, np.abs(GetRealArray('l0', number=zone)),
                                              out=np.zeros_like(lEdges), where=GetRealArray('l0', number=zone) != 0))
    }

    handler = stretchingHandlers.get(stretchingType)
    progFac, l0, dx = handler() if handler else (np.array(()), np.array(()), np.array(()))

    if stretchingType == 'combination':
        for iDim in range(3):
            if np.isclose(progFac[iDim], 0., atol=mesh_vars.tolInternal):
                continue  # Skip if factor is zero, (nElem, l0) given, factor calculated later

            progFac[iDim] = (np.abs(progFac[iDim]))**(np.sign(progFac[iDim]*l0[iDim]))

            if np.isclose(progFac[iDim], 1., atol=mesh_vars.tolInternal):
                continue  # Skip if factor is one, no stretching

            # Calculate the number of elements in the current zone
            nElems[iDim] = max(1, np.rint(np.log(1.-dx[iDim]*(1.-progFac[iDim])) / np.log(progFac[iDim])))
        print(hopout.warn(f'nElems in zone {zone} have been updated to ({nElems[0]}, {nElems[1]},{nElems[2]}).'))

    # Calculate the required factor from ratio or combination input
    if stretchingType in {'ratio', 'combination'}:
        print(hopout.warn(hopout.Colors.WARN + '─'*(46-16) + hopout.Colors.END))
        for iDim in range(3):
            if nElems[iDim] == 1 or dx[iDim] == 0:
                progFac[iDim] = 1.
                continue
            elif nElems[iDim] == 2:
                progFac[iDim] = dx[iDim] - 1.
                continue

            # Start value for Newton iteration
            progFac[iDim] = dx[iDim] / nElems[iDim]

            if np.isclose(progFac[iDim], 1., atol=mesh_vars.tolInternal):
                continue  # Skip iteration if equidistant case

            # Newton iteration
            iter = 0
            while iter < 1000:
                F  = progFac[iDim]**nElems[iDim] + dx[iDim]*(1.-progFac[iDim]) - 1.
                df = nElems[iDim]*progFac[iDim]**(nElems[iDim]-1) - dx[iDim]

                if np.isclose(F, 0., atol=mesh_vars.tolInternal) or np.isclose(df/F, 0., atol=mesh_vars.tolInternal):
                    break

                progFac[iDim] -= F / df
                iter          += 1

            if iter == 1000:
                hopout.warning('Newton iteration for computing the stretching function has failed.')
                traceback.print_stack(file=sys.stdout)
                sys.exit(1)

            progFac[iDim] = progFac[iDim]**np.sign(l0[iDim])
            print(hopout.warn(f'New stretching factor [dir {iDim}]: {progFac[iDim]}'))

        print(hopout.warn(hopout.Colors.WARN + '─'*(46-16) + hopout.Colors.END))

    # Return stretching factor
    return progFac


def TransformMesh() -> None:
    # Local imports ----------------------------------------
    from pyhope.config.config import prmfile
    from pyhope.readintools.readintools import CountOption
    from pyhope.readintools.readintools import GetReal, GetRealArray, GetStr
    from pyhope.mesh.mesh_vars import mesh
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    nMeshScale = CountOption('meshScale')
    nMeshTrans = CountOption('meshTrans')
    nMeshRot   = CountOption('meshRot')

    # Read in the mesh post-deformation flag
    meshPostDeform = GetStr('MeshPostDeform') if CountOption('MeshPostDeform') != 0 else 'none'

    # Leave if no transformation is required
    if all(x == 0 for x in [nMeshScale, nMeshTrans, nMeshRot]) and meshPostDeform == 'none':
        return None

    # Start with basic transformations
    hopout.separator()
    hopout.info('TRANSFORM MESH...')
    hopout.sep()

    hopout.routine('Performing basic transformations')
    hopout.sep()

    # Get scaling factor for mesh
    meshScale = GetReal('meshScale')

    # Get translation vector for mesh
    meshTrans = GetRealArray('meshTrans')

    # Get rotation matrix for mesh
    meshRot   = GetRealArray('meshRot')
    meshRot   = np.array(meshRot).reshape(3, 3)
    meshRotC  = GetRealArray('meshRotCenter')

    # Scale mesh
    if meshScale != 1.0:
        mesh.points *= meshScale

    # Translate mesh
    if not np.array_equal(meshTrans, [0.0, 0.0, 0.0]):
        mesh.points += meshTrans

    # Rotate mesh
    if not np.array_equal(meshRot, [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]):
        mesh.points = meshRotC + (mesh.points-meshRotC) @ meshRot

    # Exit routine if no further advanced transformation is required
    if meshPostDeform == 'none':
        hopout.sep()
        hopout.info('TRANSFORM MESH DONE!')
        return

    # Continue with advanced transformations
    hopout.sep()
    hopout.routine('Performing advanced transformations')
    hopout.routine('  Template: {}'.format(meshPostDeform))

    # Define locations of the transformation files ( Priority: prmfile folder > CWD > templates )
    DeformLocations = [
        os.path.join(os.path.dirname(prmfile), f'{meshPostDeform}.py'),                # Search folder of parameter file
        os.path.join(os.getcwd(), f'{meshPostDeform}.py'),                             # Search in CWD
        os.path.join(os.path.dirname(__file__), 'templates', f'{meshPostDeform}.py')   # Search in 'templates'
    ]

    # Check if the transformation file exists
    PostDeformMod = None
    for loc in DeformLocations:
        if os.path.exists(loc):
            spec = importlib.util.spec_from_file_location(meshPostDeform, loc)
            # Skip to the next location if spec is None
            if spec is None:
                continue

            PostDeformMod = importlib.util.module_from_spec(spec)
            sys.modules[meshPostDeform] = PostDeformMod
            spec.loader.exec_module(PostDeformMod)

            # Output filename of template
            hopout.routine('     found: {}'.format(loc))

            # Stop once the module is successfully loaded
            break

    # If the transformation file is not found, exit
    if PostDeformMod is None:
        hopout.warning(f'Post Transformation template "{meshPostDeform}" not found!')
        # Print all available default templates for post-deformation
        templist = []
        for file in os.listdir(os.path.join(os.path.dirname(__file__), 'templates')):
            if file.endswith('.py'):
                templist.append(f'  {file[:-3]}')
        hopout.warning('Available default transformation templates:' + ','.join(templist))
        sys.exit(1)

    # Perform actual post-deformation
    mesh.points = PostDeformMod.PostDeform(mesh.points)

    hopout.sep()
    hopout.info('TRANSFORM MESH DONE!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# from functools import cache
# from typing import Tuple
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


# def edgePointVTK(order: int, edge: int, node: int) -> np.ndarray:
#     match edge:
#         case 0:  # z- / base
#             return np.array([node      , 0         ], dtype=int)
#         case 1:  # y+ / base
#             return np.array([order     , node      ], dtype=int)
#         case 2:  # z+ / base
#             return np.array([order-node, order     ], dtype=int)
#         case 3:  # y- / base
#             return np.array([0         , order-node], dtype=int)
#         case _:
#             sys.exit(1)
#
#
# @cache
# def HEXMAPVTK(order: int) ->  -> Tuple[np.ndarray, np.ndarray]:
#     """ VTK -> IJK ordering for high-order hexahedrons
#         > Loosely based on [Gmsh] "generatePointsHexCGNS"
#         > [Jens Ulrich Kreber] "paraview-scripts/node_ordering.py"
#
#         > HEXTEN : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
#         > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)
#     """
#     map = np.zeros((order, order, order), dtype=int)
#
#     if order == 1:
#         map[0, 0, 0] = 0
#         tensor       = map
#         return map, tensor
#
#     # Principal vertices
#     map[0      , 0      , 0      ] = 1
#     map[order-1, 0      , 0      ] = 2
#     map[order-1, order-1, 0      ] = 3
#     map[0      , order-1, 0      ] = 4
#     map[0      , 0      , order-1] = 5
#     map[order-1, 0      , order-1] = 6
#     map[order-1, order-1, order-1] = 7
#     map[0      , order-1, order-1] = 8
#
#     if order == 2:
#         # Python indexing, 1 -> 0
#         map -= 1
#         # Reshape into 1D array, tensor-product style
#         tensor = []
#         for k in range(order):
#             for j in range(order):
#                 for i in range(order):
#                     tensor.append(int(map[i, j, k]))
#
#         return map, np.asarray(tensor)
#
#     # Internal points of base quadrangle edges (x-)
#     count = 8
#     for iFace in range(4):
#         for iNode in range(1, order-1):
#             # Assemble mapping to tuple, base quadrangle -> z = 0
#             count += 1
#             edge  = edgePointVTK(order-1, iFace, iNode)
#             index = (int(edge[0]), int(edge[1]), 0)
#             map[index] = count
#
#     # Internal points of top quadrangle edges
#     for iFace in range(4):
#         for iNode in range(1, order-1):
#             # Assemble mapping to tuple, top  quadrangle -> z = order
#             count += 1
#             edge  = edgePointVTK(order-1, iFace, iNode)
#             index = (int(edge[0]), int(edge[1]), order-1)
#             map[index] = count
#
#     # Internal points of mounting edges
#     for iFace in range(4):
#         edge  = edgePointVTK(order-1, (iFace+3) % 4, order-1)
#         for iNode in range(1, order-1):
#             # Assemble mapping to tuple, mounting edges -> z ascending
#             count += 1
#             index = (int(edge[0]), int(edge[1]), iNode)
#             map[index] = count
#     # > VTK9 swapped 3/4
#     # (0      ,0)
#     # for iNode in range(1, order-1):
#     #     # Assemble mapping to tuple, mounting edges -> z ascending
#     #     count += 1
#     #     index = (0           , 0           , iNode)
#     #     map[index] = count
#     # # (order-1,0)
#     # for iNode in range(1, order-1):
#     #     # Assemble mapping to tuple, mounting edges -> z ascending
#     #     count += 1
#     #     index = (order-1     , 0           , iNode)
#     #     map[index] = count
#     # # (order-1,order-1)
#     # for iNode in range(1, order-1):
#     #     # Assemble mapping to tuple, mounting edges -> z ascending
#     #     count += 1
#     #     index = (order-1     , order-1     , iNode)
#     #     map[index] = count
#     # # (order-1,order-1)
#     # for iNode in range(1, order-1):
#     #     # Assemble mapping to tuple, mounting edges -> z ascending
#     #     count += 1
#     #     index = (order-1     , order-1     , iNode)
#     #     map[index] = count
#
#     # WARNING:THIS IS HOW IT SHOULD BE
#     # Internal points of upstanding faces
#     # > x- face
#     k = 0
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             count += 1
#             index = (k    , i+1  , j+1  )
#             map[index] = count
#     # > x+ face
#     k = order-1
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             count += 1
#             index = (k    , i+1  , j+1  )
#             map[index] = count
#     # > y- face
#     k = 0
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             count += 1
#             index = (i+1  , k    , j+1  )
#             map[index] = count
#     # > y+ face
#     k = order-1
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             count += 1
#             index = (i+1  , k    , j+1  )
#             map[index] = count
#     # Internal points of base quadrangle (z-)
#     k = 0
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             count += 1
#             index = (i+1  , j+1  , k    )
#             map[index] = count
#
#     # Internal points of top  quadrangle (z+)
#     k = order-1
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             count += 1
#             index = (i+1  , j+1  , k    )
#             map[index] = count
#
#     # # FIXME: THIS IS HOW MESHIO GIVES IT
#     # # Internal points of base quadrangle (z-)
#     # k = 0
#     # # Fill the map
#     # for j in range(order-2):
#     #     for i in range(order-2):
#     #         count += 1
#     #         index = (i+1  , j+1  , k    )
#     #         map[index] = count
#     # # > y- face
#     # k = 0
#     # # Fill the map
#     # for j in range(order-2):
#     #     for i in range(order-2):
#     #         count += 1
#     #         index = (i+1  , k    , j+1  )
#     #         map[index] = count
#     # # > x+ face
#     # k = order-1
#     # # Fill the map
#     # for j in range(order-2):
#     #     for i in range(order-2):
#     #         count += 1
#     #         index = (k    , i+1  , j+1  )
#     #         map[index] = count
#     # # > y+ face
#     # k = order-1
#     # # Fill the map
#     # for j in range(order-2):
#     #     for i in range(order-2):
#     #         count += 1
#     #         index = (i+1  , k    , j+1  )
#     #         map[index] = count
#     # # > x- face
#     # k = 0
#     # # Fill the map
#     # for j in range(order-2):
#     #     for i in range(order-2):
#     #         count += 1
#     #         index = (k    , i+1  , j+1  )
#     #         map[index] = count
#     # # Internal points of top  quadrangle (z+)
#     # k = order-1
#     # # Fill the map
#     # for j in range(order-2):
#     #     for i in range(order-2):
#     #         count += 1
#     #         index = (i+1  , j+1  , k    )
#     #         map[index] = count
#
#     # Internal volume points as a tensor product
#     for k in range(1, order-1):
#         for j in range(1, order-1):
#             for i in range(1, order-1):
#                 count += 1
#                 index = (i  , j  , k  )
#                 map[index] = count
#
#     # Python indexing, 1 -> 0
#     map -= 1
#
#     # Reshape into 1D array, tensor-product style
#     tensor = []
#     for k in range(order):
#         for j in range(order):
#             for i in range(order):
#                 tensor.append(int(map[i, j, k]))
#
#     return map, np.asarray(tensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import sys
import traceback
from collections import defaultdict
from functools import cache
from typing import Final, Tuple, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def MeshSplitToHex(mesh: meshio.Mesh) -> meshio.Mesh:
    """ Split simplex elements into hexahedral elements

        > This routine is mostly identical to MeshChangeElemType
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.common.common_progress import ProgressBar
    from pyhope.mesh.mesh_vars import nGeo
    from pyhope.readintools.readintools import CreateLogical, GetLogical, CountOption
    # ------------------------------------------------------

    # Create logical here to avoid it showing up in the help
    CreateLogical('SplitToHex', multiple=False, default=False)

    if CountOption('doSplitToHex') == 0 and CountOption('SplitToHex') == 0:
        return mesh

    hopout.separator()
    hopout.info('SPLITTING ELEMENTS TO HEXAHEDRA...')
    hopout.sep()

    splitToHex = GetLogical('doSplitToHex') or \
                 GetLogical(  'SplitToHex')
    if not splitToHex:
        hopout.separator()
        return mesh

    # Native meshio data
    cdict: Final[dict] = mesh.cells_dict
    csets: Final[dict] = getattr(mesh, 'cell_sets', {})

    # Copy original points
    # points    = mesh.points.copy()
    points    = mesh.points
    pointl    = cast(list, mesh.points.tolist())
    elems_old = mesh.cells.copy()

    # Sanity check
    # > Check if the requested polynomial order is 1
    if nGeo > 1:
        hopout.warning('nGeo = {} not supported for element splitting'.format(nGeo))
        traceback.print_stack(file=sys.stdout)
        sys.exit(1)

    # > Check if the mesh contains any pyramids or hexahedra
    if any(s.startswith(x) for x in ['pyramid', 'hexahedron'] for s in cdict.keys()):
        unsupported = [s for s in cdict.keys() if any(s.startswith(x) for x in ['pyramid', 'hexahedron'])]
        hopout.warning('{}, are not supported for splitting, exiting...'.format(', '.join(unsupported)))
        sys.exit(1)

    faceType = ['triangle'  , 'quad'  ]
    faceNum  = [          3 ,       4 ]

    # Convert the (triangle/quad) boundary cell set into a dictionary
    csets_old = defaultdict(list)

    for cname, cblock in csets.items():
        if cblock is None:
            continue

        # Each set_blocks is a list of arrays, one entry per cell block
        for blockID, block in enumerate(cblock):
            if elems_old[blockID].type[:4] != 'quad' and elems_old[blockID].type[:8] != 'triangle':
                continue

            if block is None:
                continue

            # Sort them as a set for membership checks
            for face in block:
                nodes = cdict[elems_old[blockID].type][face]
                csets_old[frozenset(nodes)].append(cname)

    # nPoints  = len(points)
    nPoints  = len(pointl)
    nFaces   = np.zeros(2)

    # Prepare new cell blocks and new cell_sets
    elems_lst = {ftype: [] for ftype in faceType}
    csets_lst = {}

    # Hardcode quad element faces
    faceVal  = 1
    subNodes = hexa_faces()

    # Create the element sets
    ignElems:  Final[tuple] = ('vertex', 'line', 'quad', 'triangle', 'pyramid', 'hexahedron')
    meshcells: Final[tuple] = tuple((k, v) for k, v in cdict.items() if not any(x in k for x in ignElems))
    nTotalElems = sum(zdata.shape[0] for _, zdata in meshcells)
    bar = ProgressBar(value=nTotalElems, title='│             Processing Elements', length=33, threshold=1000)

    # Build an inverted index to map each node to all face keys (from csets_old) that contain it
    nodeToFace = defaultdict(set)
    for subFace in csets_old:
        for node in subFace:
            nodeToFace[node].add(subFace)

    elemSplitter = { 'tetra': (tet_to_hex_points  , tet_to_hex_split  , tet_to_hex_faces  ),
                     'wedge': (prism_to_hex_points, prism_to_hex_split, prism_to_hex_faces)}

    for cell in elems_old:
        ctype, cdata = cell.type, cell.data

        # if ctype.startswith('hexahedron'):
        #     continue

        splitPoints, splitElems, splitFaces = elemSplitter.get(ctype, (None, None, None))

        # Only process valid splits
        if splitPoints is None or splitElems is None or splitFaces is None:
            continue

        # Setup split functions
        subIdxs            = splitElems()
        oldFIdxs, subFIdxs = splitFaces()
        subPts             = splitPoints(order=1)

        # Iterate over element types
        for elem in cdata:
            # Create array for new nodes
            newPoints = tuple(np.mean(points[elem[i]], axis=0) for i in subPts)

            # Assemble the new nodes
            # newNodes   = np.concatenate((elem, np.arange(nPoints, nPoints + len(newPoints))))
            newNodes   = np.array(elem.tolist() + list(range(nPoints, nPoints + len(newPoints))), dtype=int)
            nPoints   += len(newPoints)
            pointl.extend(newPoints)

            # Reconstruct the cell sets for the boundary conditions
            # > They already exists for the triangular faces, but we create new quad faces with the edge and face centers
            # oldFaces, newFaces  = splitFaces(newNodes)
            oldFaces = tuple(frozenset(newNodes[oldFIdx]) for oldFIdx in oldFIdxs)
            newFaces = tuple(          newNodes[subFIdx]  for subFIdx in subFIdxs)  # noqa: E272

            # Deferr update for new boundary faces
            # > Instead of updating csets_old repeatedly, we collect deferred updates
            newBCFaces = []  # List of tuples: (new_face_key, combined_name)
            for oldFace, subFaces in zip(oldFaces, newFaces):
                # Use the inverted index to get candidate face keys that might contain oldFace
                candidate_sets = [nodeToFace[node] for node in oldFace if node in nodeToFace]
                if not candidate_sets:
                    continue

                # Intersection of candidate sets reduces the number of checks
                common_candidates = set.intersection(*candidate_sets)
                for candidate in common_candidates:
                    if oldFace.issubset(candidate):
                        # Combine the associated names into one string
                        combined_name = ''.join(csets_old[candidate])
                        # Create the new quadrilateral boundary faces by precomputing the frozensets
                        faceSet = [frozenset(face) for face in subFaces]
                        # This cannot be a tuple, we need to iterate over all the candidate sets
                        for key in faceSet:
                            newBCFaces.append((key, combined_name))
                        # Done with this triangular face, break out of the (inner) candidate loop
                        break

            # Process sub-elements and record new face keys with their indices
            subElems = tuple(newNodes[subIdx] for subIdx in subIdxs)
            subFaces = []  # List of tuples: (new_face_key, face_index)
            for subElem in subElems:
                # Assemble the 6 hexahedral faces
                for subNode in subNodes:
                    subFace = subElem[subNode]
                    faceSet = frozenset(subFace)
                    subFaces.append((faceSet, nFaces[faceVal]))
                    elems_lst[faceType[faceVal]].append(np.array(subFace, dtype=int))
                    nFaces[faceVal] += 1

            # Merge deferred updates with new face indices
            for newFace, faceName in newBCFaces:
                for subFace, faceIndex in subFaces:
                    if subFace == newFace:
                        csets_lst.setdefault(faceName, [[], []])
                        csets_lst[faceName][faceVal].append(faceIndex)

            # Hardcode hexahedron elements
            if 'hexahedron' not in elems_lst:
                elems_lst['hexahedron'] = []
            # Append all rows from subElems
            # elems_lst['hexahedron'].extend(np.array(subElems, dtype=int).tolist())
            elems_lst['hexahedron'].extend(subElems)

            # Update the progress bar
            bar.step()

    # Close the progress bar
    bar.close()

    # Convert lists to NumPy arrays for elems_new and csets_new
    elems_new = {}
    csets_new = {}

    for key in elems_lst:
        if   isinstance(elems_lst[key], list) and     elems_lst[key]:  # noqa: E271
            # Convert the list of accumulated arrays/lists into a single NumPy array
            elems_new[key] = np.array(elems_lst[key], dtype=int)
        elif isinstance(elems_lst[key], list) and not elems_lst[key]:
            # Determine the expected number of columns
            elems_new[key] = np.empty((0, faceNum[faceType.index(key)]), dtype=int)

    for key in csets_lst:
        csets_new[key] = tuple(np.array(lst, dtype=int) for lst in csets_lst[key])

    # Convert points_list back to a NumPy array
    points = np.array(pointl)

    mesh = meshio.Mesh(points    = points,     # noqa: E251
                       cells     = elems_new,  # noqa: E251
                       cell_sets = csets_new)  # noqa: E251
    hopout.separator()

    return mesh


@cache
def hexa_faces() -> tuple[np.ndarray, ...]:
    """ Given the 8 corner node indices of a single hexahedral element (indexed 0..7),
        return a list of new hexahedral face connectivity lists.
    """
    return (np.array((0, 1, 2, 3), dtype=int),
            np.array((4, 5, 6, 7), dtype=int),
            np.array((0, 1, 5, 4), dtype=int),
            np.array((2, 3, 7, 6), dtype=int),
            np.array((0, 3, 7, 4), dtype=int),
            np.array((1, 2, 6, 5), dtype=int),
           )


@cache
def tet_to_hex_points(order: int) -> tuple[np.ndarray, ...]:
    """
    """
    match order:
        case 1:
            return (  # Nodes on edges
                      np.array((  0,  1    ), dtype=int),           # index 4
                      np.array((  1,  2    ), dtype=int),           # index 5
                      np.array((  0,  2    ), dtype=int),           # index 6
                      np.array((  0,  3    ), dtype=int),           # index 7
                      np.array((  1,  3    ), dtype=int),           # index 8
                      np.array((  2,  3    ), dtype=int),           # index 9
                      # Nodes on faces
                      np.array((  0,  1,  2), dtype=int),           # index 10
                      np.array((  0,  1,  3), dtype=int),           # index 11
                      np.array((  1,  2,  3), dtype=int),           # index 12
                      np.array((  0,  2,  3), dtype=int),           # index 13
                      # Inside node
                      np.arange(  0,  4, dtype=int)                # index 14
                   )
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def tet_to_hex_faces() -> Tuple[list[np.ndarray], list[list[np.ndarray]]]:
    """ Given the 4 corner node indices of a single tetrahedral element (indexed 0..3),
        return the 4 triangular faces and the 12 quadrilateral faces.
    """
    # Triangular faces
    oldFaces  = [  np.array((  0,  1,  2    ), dtype=int),
                   np.array((  0,  2,  3    ), dtype=int),
                   np.array((  0,  3,  1    ), dtype=int),
                   np.array((  1,  2,  3    ), dtype=int)
                ]

    # Quadrilateral faces
    newFaces  = [  # First triangle
                 [ np.array((  0,  4,  6, 10), dtype=int),
                   np.array((  4,  1,  5, 10), dtype=int),
                   np.array((  5,  2,  6, 10), dtype=int)],
                   # Second triangle
                 [ np.array((  0,  6,  7, 13), dtype=int),
                   np.array((  6,  2,  9, 13), dtype=int),
                   np.array((  9,  3,  7, 13), dtype=int)],
                   # Third triangle
                 [ np.array((  0,  4,  7, 11), dtype=int),
                   np.array((  4,  1,  8, 11), dtype=int),
                   np.array((  8,  3,  7, 11), dtype=int)],
                   # Fourth triangle
                 [ np.array((  1,  5,  8, 12), dtype=int),
                   np.array((  5,  2,  9, 12), dtype=int),
                   np.array((  9,  3,  8, 12), dtype=int)]
                ]

    return oldFaces, newFaces


@cache
def tet_to_hex_split() -> list[np.ndarray]:
    """ Given the 4 corner node indices of a single tetrahedral element (indexed 0..3),
        return a list of new hexahedral element connectivity lists.
    """
    return [np.array((  0,  4, 10,  6,  7, 11, 14, 13), dtype=int),
            np.array((  1,  5, 10,  4,  8, 12, 14, 11), dtype=int),
            np.array((  2,  6, 10,  5,  9, 13, 14, 12), dtype=int),
            np.array((  3,  7, 13,  9,  8, 11, 14, 12), dtype=int),
           ]


@cache
def prism_to_hex_points(order: int) -> tuple[np.ndarray, ...]:
    """
    """
    match order:
        case 1:
            return (  # Nodes on edges
                      np.array((  0,  1    ), dtype=int),           # index 6
                      np.array((  1,  2    ), dtype=int),           # index 7
                      np.array((  0,  2    ), dtype=int),           # index 8
                      np.array((  3,  4    ), dtype=int),           # index 9
                      np.array((  4,  5    ), dtype=int),           # index 10
                      np.array((  3,  5    ), dtype=int),           # index 11
                      # Nodes on faces
                      np.array((  0,  1,  2), dtype=int),           # index 12
                      np.array((  3,  4,  5), dtype=int),           # index 13
                   )
        case _:
            print('Order {} not supported for element splitting'.format(order))
            traceback.print_stack(file=sys.stdout)
            sys.exit(1)


@cache
def prism_to_hex_faces() -> Tuple[list[np.ndarray], list[list[np.ndarray]]]:
    """ Given the 6 corner node indices of a single prism element (indexed 0..5),
        return the 6 -> new <- quadrilateral faces.
    """
    # Faces
    oldFaces  = [  # Triangular faces
                   np.array((  0,  1,  2    ), dtype=int),
                   np.array((  3,  4,  5    ), dtype=int),
                   # Quadrilateral faces
                   np.array((  0,  1,  4,  3), dtype=int),
                   np.array((  1,  2,  5,  4), dtype=int),
                   np.array((  2,  0,  3,  5), dtype=int),
                ]

    # Quadrilateral faces
    newFaces  = [  # First triangle
                 [ np.array((  0,  6, 12,  8), dtype=int),
                   np.array((  1,  7, 12,  6), dtype=int),
                   np.array((  2,  8, 12,  7), dtype=int)],
                   # Second triangle
                 [ np.array((  3,  9, 13, 11), dtype=int),
                   np.array((  4, 10, 13,  9), dtype=int),
                   np.array((  5, 11, 13, 10), dtype=int)],
                   # First quad face
                 [ np.array((  0,  6,  9,  3), dtype=int),
                   np.array((  6,  1,  4,  9), dtype=int)],
                   # Second quad face
                 [ np.array((  1,  7, 10,  4), dtype=int),
                   np.array((  7,  2,  5, 10), dtype=int)],
                   # Third quad face
                 [ np.array((  0,  8, 11,  3), dtype=int),
                   np.array((  8,  2,  5, 11), dtype=int)]
                ]

    return oldFaces, newFaces


@cache
def prism_to_hex_split() -> tuple[np.ndarray, ...]:
    """ Given the 6 corner node indices of a single prism element (indexed 0..5),
        return a list of new hexahedral element connectivity lists.
    """
    return (np.array((  0,  6, 12,  8,  3,  9, 13, 11), dtype=int),
            np.array((  1,  7, 12,  6,  4, 10, 13,  9), dtype=int),
            np.array((  2,  8, 12,  7,  5, 11, 13, 10), dtype=int),
           )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import gc
from typing import cast, final
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def Coords2Int(coords: np.ndarray, spacing: np.ndarray, xmin: np.ndarray) -> np.ndarray:
    """ Compute the integer discretization in each direction
    """
    disc = np.round((coords - xmin) * spacing)
    return np.asarray(disc)


def SFCResolution(kind: int, xmin: np.ndarray, xmax: np.ndarray) -> tuple[int, np.ndarray]:
    """ Compute the resolution of the SFC for the given bounding box
        and the given integer kind
    """
    blen    = xmax - xmin
    nbits   = (kind*8 - 1)  # / 3.
    intfact = cast(int, 2**nbits-1)
    spacing = np.ceil(intfact/blen)

    return np.ceil(nbits).astype(int), spacing


@final
class tBox:
    def __init__(self, mini: int, maxi: int):
        self.mini = mini
        self.intfact = 0
        self.spacing = np.zeros(3)
        self._set_bounding_box(mini, maxi)

    def _set_bounding_box(self, mini, maxi):
        blen = maxi - mini
        nbits = (np.iinfo(np.int64).bits - 1) // 3
        self.intfact = 2 ** nbits - 1
        if np.all(blen > 0):
            self.spacing = self.intfact / blen
        else:
            self.spacing = self.intfact


def SortMeshBySFC() -> None:
    # Local imports ----------------------------------------
    from hilbertcurve.hilbertcurve import HilbertCurve
    from pyhope.common.common_vars import np_mtp
    from pyhope.mesh.mesh_common import calc_elem_bary
    from pyhope.common.common_progress import ProgressBar
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    import numpy as np
    # INFO: Alternative Hilbert curve sorting (not on PyPI)
    # from hilsort import hilbert_sort
    # ------------------------------------------------------

    hopout.routine('Sorting elements along space-filling curve')

    mesh  = mesh_vars.mesh
    elems = mesh_vars.elems
    sides = mesh_vars.sides

    totalElems = len(elems)
    bar = ProgressBar(value=totalElems, title='│              Preparing Elements', length=33)

    # Global bounding box
    points = mesh.points
    xmin = np.min(points, axis=0)
    xmax = np.max(points, axis=0)

    # Calculate the element barycenters and associated element offsets
    elem_bary  = calc_elem_bary(elems)

    # Calculate the space-filling curve resolution for the given KIND
    kind = 4
    nbits, spacing = SFCResolution(kind, xmin, xmax)

    # Discretize the element positions according to the chosen resolution
    elem_disc      = Coords2Int(elem_bary, spacing, xmin)

    # Generate the space-filling curve and order elements along it
    hc             = HilbertCurve(p=nbits, n=3, n_procs=np_mtp)

    distances      = np.array(hc.distances_from_points(elem_disc))  # bottleneck
    sorted_indices = np.argsort(distances)

    # INFO: Alternative Hilbert curve sorting (not on PyPI)
    # distances      = np.array(hilbert_sort(8, elem_bary))
    # Find the new sorting with the old elem_bary
    # value_to_index = {tuple(value.tolist()): idx for idx, value in enumerate(distances)}
    # Now, create an array that maps each element to the new sorting
    # sorted_indices = np.array([value_to_index[tuple(val.tolist())] for val in elem_bary])

    # Initialize sorted cells
    sorted_elems   = tuple(elems[i] for i in sorted_indices)
    sorted_sides   = []

    bar.title('│             Processing Elements')

    # Overwrite the elem/side IDs
    offsetSide = 0
    for elemID, elem in enumerate(sorted_elems):
        elem.elemID = elemID

        for key, val in enumerate(elem.sides):
            side        = sides[val]
            side.sideID = offsetSide + key
            side.elemID = elemID

            sorted_sides.append(side)

        # Correct the sideID
        nSides      = len(elem.sides)
        elem.sides  = list(range(offsetSide, offsetSide + nSides))
        offsetSide += nSides
        bar.step()

    mesh_vars.elems = sorted_elems
    mesh_vars.sides = sorted_sides

    # Close the progress bar
    bar.close()


def SortMeshByIJK() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.mesh.mesh_common import count_elems, calc_elem_bary
    from pyhope.common.common_progress import ProgressBar
    # ------------------------------------------------------

    hopout.routine('Sorting elements along I,J,K direction')

    mesh  = mesh_vars.mesh
    elems = mesh_vars.elems
    sides = mesh_vars.sides

    # Calculate the element bary centers and type offsets
    elemBary = calc_elem_bary(elems)

    # Calculate bounding box and conversion factor
    ptp_elemBary = np.ptp(elemBary, axis=0)
    lower        = np.min(np.min(elemBary, axis=0) - 0.1 * np.min(ptp_elemBary, axis=0), axis=0)
    upper        = np.max(np.max(elemBary, axis=0) + 0.1 * np.max(ptp_elemBary, axis=0), axis=0)

    # Convert coordinates to integer space
    box       = tBox(np.floor(lower), np.ceil(upper))
    intCoords = np.rint((elemBary - box.mini) * box.spacing).astype(int)

    # Initialize lists
    nElems    = count_elems(mesh)
    nElemsIJK = np.zeros(3, dtype=int)
    structDir = np.zeros(3, dtype=bool)
    tol       = 1
    intList   = []

    for dir in range(3):
        # Sort elements by coordinate directions
        intList = intCoords[:, dir]
        sortedIndices = np.argsort(intList)
        intListSorted = intList[sortedIndices]

        # Determine structured directions
        nElems_min, nElems_max = nElems, 0
        counter = 1

        # Count the consecutive matching values to determine structure
        for iElem in range(1, nElems):
            if abs(intListSorted[iElem] - intListSorted[iElem - 1]) > tol:
                nElems_min = min(nElems_min, counter)
                nElems_max = max(nElems_max, counter)
                counter = 1
            else:
                counter += 1

        if nElems_max != nElems_min:
            nElemsIJK[dir] = 0  # Not structured
            structDir[dir] = False
        else:
            nElemsIJK[dir] = nElems_max
            structDir[dir] = True

    nStructDirs = np.sum(structDir)

    # Adjust nElemsIJK based on structured directions
    if nStructDirs == 0:
        nElemsIJK = np.array((nElems, 1, 1))
    elif nStructDirs == 1:
        structured_dir = np.argmax(structDir)
        nElemsIJK[structured_dir] = nElems // nElemsIJK[structured_dir]
        nElemsIJK[(structured_dir + 1) % 3] = nElems // nElemsIJK[structured_dir]
        nElemsIJK[(structured_dir + 2) % 3] = 1
    elif nStructDirs == 2:
        non_structured_dir = np.argmin(structDir)
        nElemsIJK[non_structured_dir] = 1
        nElemsIJK[~structDir] = nElemsIJK[~structDir][::-1]
    else:
        tIJK = np.copy(nElemsIJK)
        nElemsIJK[0] = round(np.sqrt(tIJK[1] * tIJK[2] / tIJK[0]))
        nElemsIJK[1] = round(np.sqrt(tIJK[0] * tIJK[2] / tIJK[1]))
        nElemsIJK[2] = round(np.sqrt(tIJK[0] * tIJK[1] / tIJK[2]))

    # Check for consistency in the number of elements
    if np.prod(nElemsIJK) != nElems:
        hopout.warning('Problem during sort elements by coordinate: nElems /= nElems_I * Elems_J * nElems_K')

    hopout.sep()
    hopout.info(' Number of structured dirs      : {}'.format(nStructDirs))
    hopout.info(' Number of elems [I,J,K]        : {}'.format(nElemsIJK))

    totalElems = len(elems)
    bar = ProgressBar(value=totalElems, title='│              Preparing Elements', length=33)

    # Now sort the elements based on z, y, then x coordinates
    intList        = (intCoords[:, 2] * 10000 + intCoords[:, 1]) * 10000 + intCoords[:, 0]
    sorted_indices = np.argsort(intList)

    # Initialize sorted cells
    sorted_elems   = [elems[i] for i in sorted_indices]
    sorted_sides   = []

    bar.title('│             Processing Elements')

    # Overwrite the elem/side IDs
    offsetSide = 0
    for elemID, elem in enumerate(sorted_elems):
        elem.elemID = elemID

        for key, val in enumerate(elem.sides):
            side        = sides[val]
            side.sideID = offsetSide + key
            side.elemID = elemID

            sorted_sides.append(side)

        # Correct the sideID
        nSides      = len(elem.sides)
        elem.sides  = list(range(offsetSide, offsetSide + nSides))
        offsetSide += nSides
        bar.step()

    mesh_vars.elems = sorted_elems
    mesh_vars.sides = sorted_sides

    bar.close()


def SortMesh() -> None:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import GetLogical
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('SORT MESH...')
    hopout.sep()

    mesh_vars.sortIJK = GetLogical('doSortIJK')
    hopout.sep()

    # Sort the mesh
    if mesh_vars.sortIJK:
        SortMeshByIJK()
    else:
        SortMeshBySFC()

    # Run garbage collector to release memory
    gc.collect()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
import sys
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def MeshExternal() -> meshio.Mesh:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.config.config import prmfile
    from pyhope.mesh.mesh_vars import BC
    from pyhope.mesh.reader.reader_gmsh import compatibleGMSH, ReadGMSH, BCCGNS
    from pyhope.mesh.reader.reader_gambit import ReadGambit
    from pyhope.mesh.reader.reader_hopr import ReadHOPR
    from pyhope.readintools.readintools import CountOption, GetIntArray, GetRealArray, GetStr
    # ------------------------------------------------------

    hopout.sep()
    hopout.info('LOADING EXTERNAL MESH...')

    hopout.sep()
    hopout.routine('Setting boundary conditions')
    hopout.sep()

    # Load the boundary conditions
    nBCs = CountOption('BoundaryName')
    mesh_vars.bcs = [BC() for _ in range(nBCs)]
    bcs = mesh_vars.bcs

    for iBC, bc in enumerate(bcs):
        # bc.update(name = GetStr(     'BoundaryName', number=iBC),  # noqa: E251
        #           bcid = iBC + 1,                                  # noqa: E251
        #           type = GetIntArray('BoundaryType', number=iBC))  # noqa: E251
        bc.name = GetStr(     'BoundaryName', number=iBC).lower()    # noqa: E251
        bc.bcid = iBC + 1                                            # noqa: E251
        bc.type = GetIntArray('BoundaryType', number=iBC)            # noqa: E251

    nVVs = CountOption('vv')
    mesh_vars.vvs = [dict() for _ in range(nVVs)]
    vvs = mesh_vars.vvs
    if len(vvs) > 0:
        hopout.sep()
    for iVV, _ in enumerate(vvs):
        vvs[iVV] = dict()
        vvs[iVV]['Dir'] = GetRealArray('vv', number=iVV)

    # Load the mesh(es)
    mesh   = meshio.Mesh(np.array(()), dict())
    fnames = [GetStr('Filename', number=i) for i in range(CountOption('Filename'))]

    # Check whether mesh file exists in the current directory or in the same directory
    for iFile, fname in enumerate(fnames):
        if os.path.isfile(os.path.abspath(fname)):
            fnames[iFile] = os.path.abspath(fname)
        elif os.path.isfile(os.path.join(os.path.dirname(prmfile), fname)):
            fnames[iFile] = os.path.abspath(os.path.join(os.path.dirname(prmfile), fname))
            print(hopout.warn('Mesh not found in the CWD, but found in the prmfile directory.'))
        else:
            hopout.warning('Mesh file [󰇘]/{} does not exist'.format(os.path.basename(fname)))
            sys.exit(1)

    if not all(compatibleGMSH(fname) for fname in fnames):
        if any(compatibleGMSH(fname) for fname in fnames):
            hopout.warning('Mixed file formats detected, this is untested and may not work')
            # sys.exit(1)

    # Gmsh has to come first as we cannot extend the mesh
    fgmsh = [s for s in fnames if compatibleGMSH(s)]
    if len(fgmsh) > 0:
        mesh = ReadGMSH(fgmsh)
    fnames = list(filter(lambda x: not compatibleGMSH(x), fnames))

    # Gambit meshes can extend the Gmsh mesh
    fgambit = [s for s in fnames if s.endswith('.neu')]
    if len(fgambit) > 0:
        mesh = ReadGambit(fgambit, mesh)
    fnames = list(filter(lambda x: x not in fgambit, fnames))

    # HOPR meshes can extend the Gmsh mesh
    fhopr  = [s for s in fnames if s.endswith('.h5')]
    if len(fhopr) > 0:
        mesh = ReadHOPR(fhopr, mesh)
    fnames = list(filter(lambda x: x not in fhopr, fnames))

    # If there are still files left, we have an unknown format
    if len(fnames) > 0:
        hopout.warning('Unknown file format {}, exiting...'.format(fnames))
        sys.exit(1)

    # Regenerate the boundary conditions
    if mesh_vars.CGNS.regenerate_BCs:
        mesh = BCCGNS(mesh, fgmsh)

    # Reconstruct periodicity vectors from mesh
    hasPeriodic = np.any([bcs[s].type[0] == 1 for s in range(nBCs)])
    if len(mesh_vars.vvs) == 0 and hasPeriodic:
        print(hopout.warn('Periodicity vectors neither defined in parameter file nor '
                          'in the given mesh file. Reconstructing the vectors from BCs!'))
        # Get max number of periodic alphas
        mesh_vars.vvs = [dict() for _ in range(int(np.max([np.abs(bc.type[3]) for bc in bcs])))]
        vvs = recontruct_periodicity(mesh)
        hopout.routine('The following vectors were recovered:')
        for iVV, vv in enumerate(vvs):
            hopout.printoption('vv[{}]'.format(iVV+1),'{0:}'.format(np.round(vv['Dir'],6)), 'RECOVER')
        hopout.sep()

    hopout.info('LOADING EXTERNAL MESH DONE!')
    hopout.sep()

    return mesh


def recontruct_periodicity(mesh: meshio.Mesh) -> list:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    bcs = mesh_vars.bcs
    vvs = mesh_vars.vvs

    for iVV, vv in enumerate(vvs):

        # Identify positive and negative periodic boundaries
        boundaries = {1: None, -1: None}
        for bc in [s for s in bcs if abs(s.type[3]) == iVV + 1]:
            sign = np.sign(bc.type[3])
            if boundaries[sign] is not None:
                hopout.warning("Multiple periodic boundaries found for the same direction. Exiting...")
                sys.exit(1)
            boundaries[sign] = bc.name

        # Compute mean coordinates for both boundaries as a tuple
        mean_coords = tuple(
            np.mean(mesh.points[
                np.array(sorted({
                    node for iBlock, _ in enumerate(mesh.cells)
                    if (mesh.cell_sets[bc] and mesh.cell_sets[bc][iBlock] is not None)
                    for node in mesh.cells[iBlock].data[mesh.cell_sets[bc][iBlock]].flatten()
                }))
            ], axis=0) if bc else None
            for bc in (boundaries[1], boundaries[-1])
        )

        # Store the periodicity vector if both mean coordinates exist
        if mean_coords[0] is not None and mean_coords[1] is not None:
            vv.update({"Dir": mean_coords[1] - mean_coords[0]})
        else:
            vv.update({"Dir": np.array([0.0, 0.0, 0.0])})

    return vvs
